/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing authors: Stephen Foiles (SNL), Murray Daw (SNL)
------------------------------------------------------------------------- */

#include "math.h"
#include "stdio.h"
#include "string.h"
#include "pair_eam.h"
#include "atom.h"
#include "force.h"
#include "update.h"
#include "integrate.h"
#include "comm.h"
#include "memory.h"
#include "neighbor.h"
#include "error.h"

#define MAXLINE 1024
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* ---------------------------------------------------------------------- */

PairEAM::PairEAM()
{
  style = -1;

  // set one array in each group to NULL so deallocate tests are valid

  frho_funcfl = NULL;
  mass_setfl = NULL;
  map = NULL;
  nrho_one = NULL;
  frho = NULL;
  rhor_0 = NULL;
}

/* ----------------------------------------------------------------------
   free all arrays
   check if allocated, since class can be destructed when incomplete
------------------------------------------------------------------------- */

PairEAM::~PairEAM()
{
  if (allocated) {
    memory->destroy_2d_int_array(setflag);
    memory->destroy_2d_double_array(cutsq);
  }

  if (frho_funcfl) {
    delete [] frho_funcfl;
    delete [] zr_funcfl;
    delete [] rhor_funcfl;
  }

  if (mass_setfl) {
    delete [] mass_setfl;
    memory->destroy_2d_double_array(frho_setfl);
    memory->destroy_2d_double_array(rhor_setfl);
    memory->destroy_3d_double_array(z2r_setfl);
  }

  if (map) delete [] map;

  if (nrho_one) {
    delete [] nrho_one;
    delete [] nr_one;
    delete [] drho_one;
    delete [] dr_one;
    delete [] cut_one;

    for (int i = 1; i <= ntypes_store; i++)
      if (frho_one[i]) {
	delete [] frho_one[i];
	delete [] rhor_one[i];
	delete [] zr_one[i];
      }

    delete [] frho_one;
    delete [] rhor_one;
    delete [] zr_one;
  }

  if (frho) {
    memory->destroy_2d_double_array(frho);
    memory->destroy_2d_double_array(rhor);
    memory->destroy_3d_double_array(z2r);
  }

  if (rhor_0) interpolate_deallocate();
}

/* ---------------------------------------------------------------------- */

void PairEAM::compute(int eflag, int vflag)
{
  int i,j,k,m,numneigh,itype,jtype;
  double xtmp,ytmp,ztmp,delx,dely,delz;
  double rsq,r,p,fforce,rhoip,rhojp,z2,z2p,recip,phi,phip,psip;
  int *neighs;
  double **f;

  eng_vdwl = 0.0;
  if (vflag) for (i = 0; i < 6; i++) virial[i] = 0.0;

  if (vflag == 2) f = update->integrate->f_pair;
  else f = atom->f;
  double **x = atom->x;
  double *rho = atom->rho;
  double *fp = atom->fp;
  int *type = atom->type;
  int nlocal = atom->nlocal;
  int newton_pair = force->newton_pair;

  // zero out density

  if (newton_pair) {
    m = nlocal + atom->nghost;
    for (i = 0; i < m; i++) rho[i] = 0.0;
  } else for (i = 0; i < nlocal; i++) rho[i] = 0.0;

  // compute density at each atom
  // loop over neighbors of my atoms

  for (i = 0; i < nlocal; i++) {
    xtmp = x[i][0];
    ytmp = x[i][1];
    ztmp = x[i][2];
    itype = type[i];
    neighs = neighbor->firstneigh[i];
    numneigh = neighbor->numneigh[i];

    for (k = 0; k < numneigh; k++) {
      j = neighs[k];

      delx = xtmp - x[j][0];
      dely = ytmp - x[j][1];
      delz = ztmp - x[j][2];
      rsq = delx*delx + dely*dely + delz*delz;

      if (rsq < cutforcesq) {
	jtype = type[j];
	p = sqrt(rsq)*rdr + 1.0;
	m = static_cast<int> (p);
	m = MIN(m,nr-1);
	p -= m;
	p = MIN(p,1.0);
	rho[i] += ((rhor_3[jtype][m]*p + rhor_2[jtype][m])*p + 
		   rhor_1[jtype][m])*p + rhor_0[jtype][m];
	if (newton_pair || j < nlocal)
	  rho[j] += ((rhor_3[itype][m]*p + rhor_2[itype][m])*p + 
		     rhor_1[itype][m])*p + rhor_0[itype][m];
      }
    }
  }

  // communicate and sum densities

  if (newton_pair) comm->reverse_comm_rho();

  // compute embedding function
  // sum full phi to eng_vdwl since this is per-atom embedding enery

  for (i = 0; i < nlocal; i++) {
    p = rho[i]*rdrho + 1.0;
    m = static_cast<int> (p);
    m = MAX(1,MIN(m,nrho-1));
    p -= m;
    fp[i] = (frho_6[itype][m]*p + frho_5[itype][m])*p + frho_4[itype][m];
    if (eflag) {
      phi = ((frho_3[itype][m]*p + frho_2[itype][m])*p + 
	     frho_1[itype][m])*p + frho_0[itype][m];
      eng_vdwl += phi;
    }
  }

  // communicate embedding function

  comm->comm_embed();
      
  // compute forces on each atom
  // loop over neighbors of my atoms

  for (i = 0; i < nlocal; i++) {
    xtmp = x[i][0];
    ytmp = x[i][1];
    ztmp = x[i][2];
    itype = type[i];
    neighs = neighbor->firstneigh[i];
    numneigh = neighbor->numneigh[i];

    for (k = 0; k < numneigh; k++) {
      j = neighs[k];

      delx = xtmp - x[j][0];
      dely = ytmp - x[j][1];
      delz = ztmp - x[j][2];
      rsq = delx*delx + dely*dely + delz*delz;

      if (rsq < cutforcesq) {
	jtype = type[j];
	r = sqrt(rsq);
	p = r*rdr + 1.0;
	m = static_cast<int> (p);
	m = MIN(m,nr-1);
	p -= m;
	p = MIN(p,1.0);
	rhoip = (rhor_6[itype][m]*p + rhor_5[itype][m])*p + 
	  rhor_4[itype][m];
	rhojp = (rhor_6[jtype][m]*p + rhor_5[jtype][m])*p + 
	  rhor_4[jtype][m];
	z2 = ((z2r_3[itype][jtype][m]*p + z2r_2[itype][jtype][m])*p + 
	      z2r_1[itype][jtype][m])*p + z2r_0[itype][jtype][m];
	z2p = (z2r_6[itype][jtype][m]*p + z2r_5[itype][jtype][m])*p + 
	  z2r_4[itype][jtype][m];
	recip = 1.0/r;
	phi = z2*recip;
	phip = z2p*recip - phi*recip;
	psip = fp[i]*rhojp + fp[j]*rhoip + phip;
	fforce = psip*recip;

	f[i][0] -= delx*fforce;
	f[i][1] -= dely*fforce;
	f[i][2] -= delz*fforce;
	if (newton_pair || j < nlocal) {
	  f[j][0] += delx*fforce;
	  f[j][1] += dely*fforce;
	  f[j][2] += delz*fforce;
	}

	if (eflag) {
	  if (newton_pair || j < nlocal) eng_vdwl += phi;
	  else eng_vdwl += 0.5*phi;
	}

	if (vflag == 1) {
	  if (newton_pair || j < nlocal) {
	    virial[0] -= delx*delx*fforce;
	    virial[1] -= dely*dely*fforce;
	    virial[2] -= delz*delz*fforce;
	    virial[3] -= delx*dely*fforce;
	    virial[4] -= delx*delz*fforce;
	    virial[5] -= dely*delz*fforce;
	  } else {
	    virial[0] -= 0.5*delx*delx*fforce;
	    virial[1] -= 0.5*dely*dely*fforce;
	    virial[2] -= 0.5*delz*delz*fforce;
	    virial[3] -= 0.5*delx*dely*fforce;
	    virial[4] -= 0.5*delx*delz*fforce;
	    virial[5] -= 0.5*dely*delz*fforce;
	  }
	}
      }
    }
  }
  if (vflag == 2) virial_compute();
}

/* ----------------------------------------------------------------------
   allocate all arrays
------------------------------------------------------------------------- */

void PairEAM::allocate()
{
  allocated = 1;
  int n = atom->ntypes;

  setflag = memory->create_2d_int_array(n+1,n+1,"pair:setflag");
  for (int i = 1; i <= n; i++)
    for (int j = i; j <= n; j++)
      setflag[i][j] = 0;

  cutsq = memory->create_2d_double_array(n+1,n+1,"pair:cutsq");
}

/* ----------------------------------------------------------------------
   global settings
------------------------------------------------------------------------- */

void PairEAM::settings(int narg, char **arg)
{
  if (narg > 0) error->all("Incorrect args in pair_style command");
}

/* ----------------------------------------------------------------------
   set i,j coeffs from one line in input script
------------------------------------------------------------------------- */

void PairEAM::set_coeff(int narg, char **arg)
{
  if (narg < 2) error->all("Incorrect args in pair_coeff command");
  if (!allocated) allocate();

  // parse pair of atom types

  int ilo,ihi,jlo,jhi;
  force->bounds(arg[0],atom->ntypes,ilo,ihi);
  force->bounds(arg[1],atom->ntypes,jlo,jhi);

  // EAM funcfl file
  // if 1st time, allocate memory for all atom types via pre_funcfl
  // read file via read_funcfl
  // for each i,i type pair, store read-in coeffs in *_one via store_funcfl

  if (narg == 3) {
    if (style == 0) error->all("Cannot mix setfl/funcfl EAM files\n");
    if (style == -1) pre_funcfl();
    style = 1;

    read_funcfl(arg[2]);

    int count = 0;
    for (int i = ilo; i <= ihi; i++) {
      for (int j = MAX(jlo,i); j <= jhi; j++) {
	if (i == j) {
	  atom->set_mass(i,mass_funcfl);
	  store_funcfl(i);
	  setflag[i][i] = 2;
	  count++;
	}
      }
    }

    if (count == 0) error->all("Incorrect args in pair_coeff command");

    // EAM setfl file
    // check that atom types are * *
    // read file via read_setfl
    // create map[i] = file type corresponding to atom type i
    
  } else if (narg == 3 + atom->ntypes) {
    if (style == 1) error->all("Cannot mix setfl/funcfl EAM files\n");
    style = 0;

    if (strcmp(arg[0],"*") || strcmp(arg[1],"*"))
      error->all("Must use atom types * * with EAM setfl files");

    read_setfl(arg[2]);

    if (map) delete [] map;
    map = new int[atom->ntypes+1];

    for (int i = 1; i <= atom->ntypes; i++) {
      int imap = atoi(arg[i+2]);
      if (imap < 1 || imap > ntypes_setfl)
	error->all("Illegal atom type in EAM setfl command");
      atom->set_mass(i,mass_setfl[imap]);
      map[i] = imap;
      setflag[i][i] = 2;
    }

  } else error->all("Incorrect args in pair_coeff command");
}

/* ----------------------------------------------------------------------
   set i,i coeffs from one line in read_data file
------------------------------------------------------------------------- */

void PairEAM::set_coeff(char *str)
{
  error->all("Cannot setup EAM potentials from data file");
}

/* ----------------------------------------------------------------------
   error check and initialize all values needed for force computation
------------------------------------------------------------------------- */

void PairEAM::init()
{
  int i,j;

  // all i,i terms must be set at this point

  if (!allocated) error->all("Pair coeffs are not set");
  for (i = 1; i <= atom->ntypes; i++)
    if (setflag[i][i] == 0) error->all("All pair coeffs are not set");

  // setup multi-type setfl arrays
  // for funcfl files, convert single-type arrays to mutli-type setfl format
  // for setfl files, map read-in types to atom types

  if (style == 1) convert_funcfl();
  else if (style == 0) {
    nrho = nrho_setfl;
    nr = nr_setfl;
    drho = drho_setfl;
    dr = dr_setfl;
    cutforce = cut_setfl;

    pre_setfl();

    for (i = 1; i <= atom->ntypes; i++)
      for (j = 1; j <= atom->ntypes; j++)
	store_setfl(i,j,map[i],map[j]);
  }

  // create spline coeffs

  interpolate();

  // set cutsq to same global cutforce for each i,j pair

  for (i = 1; i <= atom->ntypes; i++)
    for (j = 1; j <= atom->ntypes; j++)
      cutsq[i][j] = cutforce*cutforce;

  cutforcesq = cutforce*cutforce;
}

/* ----------------------------------------------------------------------
   allocate arrays for storing funcfl data for each atom type
   no need to check for deallocation of nrho_one, etc
     since this function can only be called once
------------------------------------------------------------------------- */

void PairEAM::pre_funcfl()
{
  int n = atom->ntypes;

  nrho_one = new int[n+1];
  nr_one = new int[n+1];
  drho_one = new double[n+1];
  dr_one = new double[n+1];
  cut_one = new double[n+1];

  typedef double *dptr;
  frho_one = new dptr[n+1];
  rhor_one = new dptr[n+1];
  zr_one = new dptr[n+1];

  // set NULL ptrs for each type, so later deallocation tests will be valid

  for (int i = 1; i <= n; i++) {
    frho_one[i] = NULL;
    rhor_one[i] = NULL;
    zr_one[i] = NULL;
  }

  // store ntypes so can deallocate in destructor even after atom is gone

  ntypes_store = atom->ntypes;
}

/* ----------------------------------------------------------------------
   read a single element EAM potential file and bcast values
   sets the *_funcfl variables in the pair_eam.h file
------------------------------------------------------------------------- */

void PairEAM::read_funcfl(char *file)
{
  FILE *fp;
  char line[MAXLINE];

  // free memory from previous funcfl file

  if (frho_funcfl) {
    delete [] frho_funcfl;
    delete [] zr_funcfl;
    delete [] rhor_funcfl;
  }

  // open potential file

  int me = comm->me;
  if (me == 0) {
    fp = fopen(file,"r");
    if (fp == NULL) {
      char str[128];
      sprintf(str,"Cannot open EAM potential file %s",file);
      error->one(str);
    }
  }

  // read header

  int tmp;

  if (me == 0) {
    fgets(line,MAXLINE,fp);
    fgets(line,MAXLINE,fp);
    sscanf(line,"%d %lg",&tmp,&mass_funcfl);
    fgets(line,MAXLINE,fp);
    sscanf(line,"%d %lg %d %lg %lg",&nrho_funcfl,&drho_funcfl,
	   &nr_funcfl,&dr_funcfl,&cut_funcfl);
  }

  MPI_Bcast(&mass_funcfl,1,MPI_DOUBLE,0,world);
  MPI_Bcast(&nrho_funcfl,1,MPI_INT,0,world);
  MPI_Bcast(&drho_funcfl,1,MPI_DOUBLE,0,world);
  MPI_Bcast(&nr_funcfl,1,MPI_INT,0,world);
  MPI_Bcast(&dr_funcfl,1,MPI_DOUBLE,0,world);
  MPI_Bcast(&cut_funcfl,1,MPI_DOUBLE,0,world);

  // allocate input arrays and read them

  frho_funcfl = new double[nrho_funcfl+1];
  zr_funcfl = new double[nr_funcfl+1];
  rhor_funcfl = new double[nr_funcfl+1];

  if (me == 0) grab(fp,nrho_funcfl,&frho_funcfl[1]);
  MPI_Bcast(&frho_funcfl[1],nrho_funcfl,MPI_DOUBLE,0,world);

  if (me == 0) grab(fp,nr_funcfl,&zr_funcfl[1]);
  MPI_Bcast(&zr_funcfl[1],nr_funcfl,MPI_DOUBLE,0,world);

  if (me == 0) grab(fp,nr_funcfl,&rhor_funcfl[1]);
  MPI_Bcast(&rhor_funcfl[1],nr_funcfl,MPI_DOUBLE,0,world);

  // close the potential file

  if (me == 0) fclose(fp);
}

/* ----------------------------------------------------------------------
   store read-in funcfl values in funcfl arrays
------------------------------------------------------------------------- */

void PairEAM::store_funcfl(int itype)
{
  // free memory from previous funcfl storage

  if (frho_one[itype]) {
    delete [] frho_one[itype];
    delete [] rhor_one[itype];
    delete [] zr_one[itype];
  }

  nrho_one[itype] = nrho_funcfl;
  nr_one[itype] = nr_funcfl;
  drho_one[itype] = drho_funcfl;
  dr_one[itype] = dr_funcfl;
  cut_one[itype] = cut_funcfl;

  frho_one[itype] = new double[nrho_funcfl+1];
  rhor_one[itype] = new double[nr_funcfl+1];
  zr_one[itype] = new double[nr_funcfl+1];

  for (int i = 1; i <= nrho_funcfl; i++) frho_one[itype][i] = frho_funcfl[i];
  for (int i = 1; i <= nr_funcfl; i++) rhor_one[itype][i] = rhor_funcfl[i];
  for (int i = 1; i <= nr_funcfl; i++) zr_one[itype][i] = zr_funcfl[i];
}

/* ----------------------------------------------------------------------
   read a multi-element alloy EAM potential file and bcast values
   sets the *_setfl variables in the pair_eam.h file
------------------------------------------------------------------------- */

void PairEAM::read_setfl(char *file)
{
  int i,j,m;
  FILE *fp;
  char line[MAXLINE];

  // free memory from previous setfl file

  if (mass_setfl) {
    delete [] mass_setfl;
    memory->destroy_2d_double_array(frho_setfl);
    memory->destroy_2d_double_array(rhor_setfl);
    memory->destroy_3d_double_array(z2r_setfl);
  }

  // open the potential file

  int me = comm->me;
  if (me == 0) {
    fp = fopen(file,"r");
    if (fp == NULL) {
      char str[128];
      sprintf(str,"Cannot open EAM potential file %s",file);
      error->one(str);
    }
  }

  // read header

  if (me == 0) {
    fgets(line,MAXLINE,fp);
    fgets(line,MAXLINE,fp);
    fgets(line,MAXLINE,fp);
    fgets(line,MAXLINE,fp);
    sscanf(line,"%d",&ntypes_setfl);
    fgets(line,MAXLINE,fp);
    sscanf(line,"%d %lg %d %lg %lg",&nrho_setfl,&drho_setfl,&nr_setfl,
	   &dr_setfl,&cut_setfl);
  }

  MPI_Bcast(&ntypes_setfl,1,MPI_INT,0,world);
  MPI_Bcast(&nrho_setfl,1,MPI_INT,0,world);
  MPI_Bcast(&drho_setfl,1,MPI_DOUBLE,0,world);
  MPI_Bcast(&nr_setfl,1,MPI_INT,0,world);
  MPI_Bcast(&dr_setfl,1,MPI_DOUBLE,0,world);
  MPI_Bcast(&cut_setfl,1,MPI_DOUBLE,0,world);

  // allocate input arrays

  mass_setfl = new double[ntypes_setfl+1];
  frho_setfl = memory->create_2d_double_array(ntypes_setfl+1,nrho_setfl+1,
					      "eam:frho_setfl");
  rhor_setfl = memory->create_2d_double_array(ntypes_setfl+1,nr_setfl+1,
					      "eam:rhor_setfl");
  z2r_setfl = memory->create_3d_double_array(ntypes_setfl+1,ntypes_setfl+1,
					     nr_setfl+1,"eam:z2r_setfl");

  // read 3 arrays
  // frho and rhor are one per type
  // read z2r for each i <= j pair, copy values for i > j

  int tmp;

  for (i = 1; i <= ntypes_setfl; i++) {
    if (me == 0) {
      fgets(line,MAXLINE,fp);
      sscanf(line,"%d %lg",&tmp,&mass_setfl[i]);
    }
    MPI_Bcast(&mass_setfl[i],1,MPI_DOUBLE,0,world);

    if (me == 0) grab(fp,nrho_setfl,&frho_setfl[i][1]);
    MPI_Bcast(&frho_setfl[i][1],nrho_setfl,MPI_DOUBLE,0,world);
    if (me == 0) grab(fp,nr_setfl,&rhor_setfl[i][1]);
    MPI_Bcast(&rhor_setfl[i][1],nr_setfl,MPI_DOUBLE,0,world);
  }
  
  for (i = 1; i <= ntypes_setfl; i++) {
    for (j = 1; j <= i; j++) {
      if (me == 0) grab(fp,nr_setfl,&z2r_setfl[i][j][1]);
      MPI_Bcast(&z2r_setfl[i][j][1],nr_setfl,MPI_DOUBLE,0,world);
    }
  }

  for (i = 1; i <= ntypes_setfl; i++)
    for (j = i+1; j <= ntypes_setfl; j++)
      for (m = 1; m <= nr_setfl; m++) z2r_setfl[i][j][m] = z2r_setfl[j][i][m];

  // close the potential file

  if (me == 0) fclose(fp);
}

/* ----------------------------------------------------------------------
   allocate arrays for storing final multi-type setfl data
------------------------------------------------------------------------- */

void PairEAM::pre_setfl()
{
  // free memory from previous allocation

  if (frho) {
    memory->destroy_2d_double_array(frho);
    memory->destroy_2d_double_array(rhor);
    memory->destroy_3d_double_array(z2r);
  }

  int n = atom->ntypes;

  frho = memory->create_2d_double_array(n+1,nrho+1,"eam:frho");
  rhor = memory->create_2d_double_array(n+1,nr+1,"eam:rhor");
  z2r = memory->create_3d_double_array(n+1,n+1,nr+1,"eam:z2r");
}

/* ----------------------------------------------------------------------
   store read-in setfl values in setfl arrays
------------------------------------------------------------------------- */

void PairEAM::store_setfl(int itype, int jtype, int imap, int jmap)
{
  int m;

  if (itype == jtype) {
    for (m = 1; m <= nrho; m++) frho[itype][m] = frho_setfl[imap][m];
    for (m = 1; m <= nr; m++) rhor[itype][m] = rhor_setfl[imap][m];
  }
  for (m = 1; m <= nr; m++) z2r[itype][jtype][m] = z2r_setfl[imap][jmap][m];
}

/* ----------------------------------------------------------------------
   convert list of funcfl potentials to setfl format
------------------------------------------------------------------------- */

void PairEAM::convert_funcfl()
{
  int ntypes = atom->ntypes;

  // determine max values for all types

  dr = dr_one[1];
  drho = drho_one[1];
  double rmax = (nr_one[1]-1) * dr_one[1];
  double rhomax = (nrho_one[1]-1) * drho_one[1];
  cutforce = cut_one[1];

  for (int i = 2; i <= atom->ntypes; i++) {
    dr = MAX(dr,dr_one[i]);
    drho = MAX(drho,drho_one[i]);
    rmax = MAX(rmax,(nr_one[i]-1) * dr_one[i]);
    rhomax = MAX(rhomax,(nrho_one[i]-1) * drho_one[i]);
    cutforce = MAX(cutforce,cut_one[i]);
  }

  nr = static_cast<int> (rmax/dr + 0.5);
  nrho = static_cast<int> (rhomax/drho + 0.5);

  // allocate temporary array for zr2 values

  double **zrtmp = memory->create_2d_double_array(ntypes+1,nr+1,"eam:zrtmp");

  // allocate permanent arrays

  pre_setfl();

  // convert input arrays to potential arrays
  // map to a single grid with a single cutoff

  int i,j,k,m; 
  double r,p,cof1,cof2,cof3,cof4;
  double two = 2.0;
  
  for (i = 1; i <= ntypes; i++) {
    for (m = 1; m <= nrho; m++) {
      r = (m-1)*drho;
      p = r/drho_one[i] + 1.0;
      k = static_cast<int> (p);
      k = MIN(k,nrho_one[i]-2);
      k = MAX(k,2);
      p -= k;
      p = MIN(p,two);
      cof1 = -0.166666667*p*(p-1.0)*(p-2.0);
      cof2 = 0.5*(p*p-1.0)*(p-2.0);
      cof3 = -0.5*p*(p+1.0)*(p-2.0);
      cof4 = 0.166666667*p*(p*p-1.0);
      frho[i][m] = cof1*frho_one[i][k-1] + cof2*frho_one[i][k] + 
	cof3*frho_one[i][k+1] + cof4*frho_one[i][k+2];
    }
  }

  for (i = 1; i <= ntypes; i++) {
    for (m = 1; m <= nr; m++) {
      r = (m-1)*dr;
      p = r/dr_one[i] + 1.0;
      k = static_cast<int> (p);
      k = MIN(k,nr_one[i]-2);
      k = MAX(k,2);
      p -= k;
      p = MIN(p,two);
      cof1 = -0.166666667*p*(p-1.0)*(p-2.0);
      cof2 = 0.5*(p*p-1.0)*(p-2.0);
      cof3 = -0.5*p*(p+1.0)*(p-2.0);
      cof4 = 0.166666667*p*(p*p-1.0);
      rhor[i][m] = cof1*rhor_one[i][k-1] + cof2*rhor_one[i][k] +
	cof3*rhor_one[i][k+1] + cof4*rhor_one[i][k+2];
      zrtmp[i][m] = cof1*zr_one[i][k-1] + cof2*zr_one[i][k] +
	cof3*zr_one[i][k+1] + cof4*zr_one[i][k+2];
    }
  }

  // i,j terms for z2r

  for (i = 1; i <= ntypes; i++)
    for (j = 1; j <= ntypes; j++)
      for (m = 1; m <= nr; m++)
	z2r[i][j][m] = 27.2*0.529*zrtmp[i][m]*zrtmp[j][m];

  // free local memory

  memory->destroy_2d_double_array(zrtmp);
}

/* ----------------------------------------------------------------------
   interpolate EAM potentials
------------------------------------------------------------------------- */

void PairEAM::interpolate()
{
  int i,j,m;

  // free memory from previous interpolation

  if (rhor_0) interpolate_deallocate();

  // interpolation settings

  rdr = 1.0/dr;
  rdrho = 1.0/drho;

  // allocate coeff arrays

  int n = atom->ntypes;

  rhor_0 =
    memory->create_2d_double_array(n+1,nr+1*sizeof(double),"eam:rhor_0");
  rhor_1 =
    memory->create_2d_double_array(n+1,nr+1*sizeof(double),"eam:rhor_1");
  rhor_2 =
    memory->create_2d_double_array(n+1,nr+1*sizeof(double),"eam:rhor_2");
  rhor_3 =
    memory->create_2d_double_array(n+1,nr+1*sizeof(double),"eam:rhor_3");
  rhor_4 =
    memory->create_2d_double_array(n+1,nr+1*sizeof(double),"eam:rhor_4");
  rhor_5 =
    memory->create_2d_double_array(n+1,nr+1*sizeof(double),"eam:rhor_5");
  rhor_6 =
    memory->create_2d_double_array(n+1,nr+1*sizeof(double),"eam:rhor_6");

  frho_0 = 
    memory->create_2d_double_array(n+1,nrho+1*sizeof(double),"eam:frho_0");
  frho_1 =
    memory->create_2d_double_array(n+1,nrho+1*sizeof(double),"eam:frho_1");
  frho_2 =
    memory->create_2d_double_array(n+1,nrho+1*sizeof(double),"eam:frho_2");
  frho_3 =
    memory->create_2d_double_array(n+1,nrho+1*sizeof(double),"eam:frho_3");
  frho_4 =
    memory->create_2d_double_array(n+1,nrho+1*sizeof(double),"eam:frho_4");
  frho_5 =
    memory->create_2d_double_array(n+1,nrho+1*sizeof(double),"eam:frho_5");
  frho_6 =
    memory->create_2d_double_array(n+1,nrho+1*sizeof(double),"eam:frho_6");

  z2r_0 = 
    memory->create_3d_double_array(n+1,n+1,nr+1*sizeof(double),"eam:z2r_0");
  z2r_1 = 
    memory->create_3d_double_array(n+1,n+1,nr+1*sizeof(double),"eam:z2r_1");
  z2r_2 = 
    memory->create_3d_double_array(n+1,n+1,nr+1*sizeof(double),"eam:z2r_2");
  z2r_3 = 
    memory->create_3d_double_array(n+1,n+1,nr+1*sizeof(double),"eam:z2r_3");
  z2r_4 = 
    memory->create_3d_double_array(n+1,n+1,nr+1*sizeof(double),"eam:z2r_4");
  z2r_5 = 
    memory->create_3d_double_array(n+1,n+1,nr+1*sizeof(double),"eam:z2r_5");
  z2r_6 = 
    memory->create_3d_double_array(n+1,n+1,nr+1*sizeof(double),"eam:z2r_6");

  // frho interpolation

  for (i = 1; i <= atom->ntypes; i++) {
    for (m = 1; m <= nrho; m++) frho_0[i][m] = frho[i][m];

    frho_1[i][1] = frho_0[i][2]-frho_0[i][1];
    frho_1[i][2] = 0.5*(frho_0[i][3]-frho_0[i][1]);
    frho_1[i][nrho-1] = 0.5*(frho_0[i][nrho]-frho_0[i][nrho-2]);
    frho_1[i][nrho] = frho_0[i][nrho]-frho_0[i][nrho-1];

    for (m = 3; m <= nrho-2; m++)
      frho_1[i][m] = ((frho_0[i][m-2]-frho_0[i][m+2]) + 
		       8.0*(frho_0[i][m+1]-frho_0[i][m-1]))/12.0;

    for (m = 1; m <= nrho-1; m++) {
      frho_2[i][m] = 3.*(frho_0[i][m+1]-frho_0[i][m]) - 
	2.0*frho_1[i][m] - frho_1[i][m+1];
      frho_3[i][m] = frho_1[i][m] + frho_1[i][m+1] - 
	2.0*(frho_0[i][m+1]-frho_0[i][m]);
    }

    frho_2[i][nrho] = 0.0;
    frho_3[i][nrho] = 0.0;

    for (m = 1; m <= nrho; m++) {
      frho_4[i][m] = frho_1[i][m]/drho;
      frho_5[i][m] = 2.0*frho_2[i][m]/drho;
      frho_6[i][m] = 3.0*frho_3[i][m]/drho;
    }
  }

  // rhor interpolation

  for (i = 1; i <= atom->ntypes; i++) {
    for (m = 1; m <= nr; m++) rhor_0[i][m] = rhor[i][m];

    rhor_1[i][1] = rhor_0[i][2]-rhor_0[i][1];
    rhor_1[i][2] = 0.5*(rhor_0[i][3]-rhor_0[i][1]);
    rhor_1[i][nr-1] = 0.5*(rhor_0[i][nr]-rhor_0[i][nr-2]);
    rhor_1[i][nr] = 0.0;

    for (m = 3; m <= nr-2; m++)
      rhor_1[i][m] = ((rhor_0[i][m-2]-rhor_0[i][m+2]) + 
		       8.0*(rhor_0[i][m+1]-rhor_0[i][m-1]))/12.;

    for (m = 1; m <= nr-1; m++) {
      rhor_2[i][m] = 3.0*(rhor_0[i][m+1]-rhor_0[i][m]) - 
	2.0*rhor_1[i][m] - rhor_1[i][m+1];
      rhor_3[i][m] = rhor_1[i][m] + rhor_1[i][m+1] - 
	2.0*(rhor_0[i][m+1]-rhor_0[i][m]);
    }

    rhor_2[i][nr] = 0.0;
    rhor_3[i][nr] = 0.0;

    for (m = 1; m <= nr; m++) {
      rhor_4[i][m] = rhor_1[i][m]/dr;
      rhor_5[i][m] = 2.0*rhor_2[i][m]/dr;
      rhor_6[i][m] = 3.0*rhor_3[i][m]/dr;
    }
  }

  // z2r interpolation

  for (i = 1; i <= atom->ntypes; i++) {
    for (j = 1; j <= atom->ntypes; j++) {
      for (m = 1; m <= nr; m++) z2r_0[i][j][m] = z2r[i][j][m];

      z2r_1[i][j][1] = z2r_0[i][j][2]-z2r_0[i][j][1];
      z2r_1[i][j][2] = 0.5*(z2r_0[i][j][3]-z2r_0[i][j][1]);
      z2r_1[i][j][nr-1] = 0.5*(z2r_0[i][j][nr]-z2r_0[i][j][nr-2]);
      z2r_1[i][j][nr] = 0.0;

      for (m = 3; m <= nr-2; m++) 
	z2r_1[i][j][m] = ((z2r_0[i][j][m-2]-z2r_0[i][j][m+2]) + 
			   8.0*(z2r_0[i][j][m+1]-z2r_0[i][j][m-1]))/12.;

      for (m = 1; m <= nr-1; m++) {
	z2r_2[i][j][m] = 3.0*(z2r_0[i][j][m+1]-z2r_0[i][j][m]) - 
	  2.0*z2r_1[i][j][m] - z2r_1[i][j][m+1];
	z2r_3[i][j][m] = z2r_1[i][j][m] + z2r_1[i][j][m+1] - 
	  2.0*(z2r_0[i][j][m+1]-z2r_0[i][j][m]);
      }

      z2r_2[i][j][nr] = 0.0;
      z2r_3[i][j][nr] = 0.0;

      for (m = 1; m <= nr; m++) {
	z2r_4[i][j][m] = z2r_1[i][j][m]/dr;
	z2r_5[i][j][m] = 2.0*z2r_2[i][j][m]/dr;
	z2r_6[i][j][m] = 3.0*z2r_3[i][j][m]/dr;
      }
    }
  }
}

/* ----------------------------------------------------------------------
   grab n values from file fp and put them in list
   values can be several to a line
   only called by proc 0
------------------------------------------------------------------------- */

void PairEAM::grab(FILE *fp, int n, double *list)
{
  char *ptr;
  char line[MAXLINE];

  int i = 0;
  while (i < n) {
    fgets(line,MAXLINE,fp);
    ptr = strtok(line," \t\n");
    list[i++] = atof(ptr);
    while (ptr = strtok(NULL," \t\n")) list[i++] = atof(ptr);
  }
}

/* ----------------------------------------------------------------------
   deallocate spline interpolation arrays
------------------------------------------------------------------------- */

void PairEAM::interpolate_deallocate()
{
  memory->destroy_2d_double_array(rhor_0);
  memory->destroy_2d_double_array(rhor_1);
  memory->destroy_2d_double_array(rhor_2);
  memory->destroy_2d_double_array(rhor_3);
  memory->destroy_2d_double_array(rhor_4);
  memory->destroy_2d_double_array(rhor_5);
  memory->destroy_2d_double_array(rhor_6);

  memory->destroy_2d_double_array(frho_0);
  memory->destroy_2d_double_array(frho_1);
  memory->destroy_2d_double_array(frho_2);
  memory->destroy_2d_double_array(frho_3);
  memory->destroy_2d_double_array(frho_4);
  memory->destroy_2d_double_array(frho_5);
  memory->destroy_2d_double_array(frho_6);

  memory->destroy_3d_double_array(z2r_0);
  memory->destroy_3d_double_array(z2r_1);
  memory->destroy_3d_double_array(z2r_2);
  memory->destroy_3d_double_array(z2r_3);
  memory->destroy_3d_double_array(z2r_4);
  memory->destroy_3d_double_array(z2r_5);
  memory->destroy_3d_double_array(z2r_6);
}
