/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "mpi.h"
#include "atom_angle.h"
#include "domain.h"
#include "force.h"
#include "modify.h"
#include "fix.h"
#include "output.h"
#include "error.h"
#include "memory.h"
#include "comm.h"

#define DELTA 10000
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* ---------------------------------------------------------------------- */

AtomAngle::AtomAngle(char *style_in) : Atom(style_in)
{
  molecular = 1;
  bonds_allow = angles_allow = 1;
  dihedrals_allow = impropers_allow = 0;
  mass_require = 1;
  dipole_require = 0;

  size_comm = 3;
  size_reverse = 3;
  size_border = 7;
  size_atom_valid = 6;

  tag = type = mask = image = NULL;
  x = v = f = NULL;

  molecule = NULL;

  num_bond = NULL;
  bond_type = bond_atom = NULL;

  num_angle = NULL;
  angle_type = angle_atom1 = angle_atom2 = angle_atom3 = NULL;

  maxspecial = 1;
  nspecial = NULL;
  special = NULL;
}

/* ---------------------------------------------------------------------- */

AtomAngle::~AtomAngle()
{
  memory->sfree(tag);
  memory->sfree(type);
  memory->sfree(mask);
  memory->sfree(image);
  memory->destroy_2d_double_array(x);
  memory->destroy_2d_double_array(v);
  memory->destroy_2d_double_array(f);

  memory->sfree(molecule);

  memory->sfree(num_bond);
  memory->destroy_2d_int_array(bond_type);
  memory->destroy_2d_int_array(bond_atom);

  memory->sfree(num_angle);
  memory->destroy_2d_int_array(angle_type);
  memory->destroy_2d_int_array(angle_atom1);
  memory->destroy_2d_int_array(angle_atom2);
  memory->destroy_2d_int_array(angle_atom3);

  memory->destroy_2d_int_array(nspecial);
  memory->destroy_2d_int_array(special);
}

/* ----------------------------------------------------------------------
   grow atom arrays
   n = 0 grows arrays by DELTA
   n > 0 allocates arrays to size n 
------------------------------------------------------------------------- */

void AtomAngle::grow(int n)
{
  if (n == 0) nmax += DELTA;
  else nmax = n;

  tag = (int *) memory->srealloc(tag,nmax*sizeof(int),"atom:tag");
  type = (int *) memory->srealloc(type,nmax*sizeof(int),"atom:type");
  mask = (int *) memory->srealloc(mask,nmax*sizeof(int),"atom:mask");
  image = (int *) memory->srealloc(image,nmax*sizeof(int),"atom:image");
  x = memory->grow_2d_double_array(x,nmax,3,"atom:x");
  v = memory->grow_2d_double_array(v,nmax,3,"atom:v");
  f = memory->grow_2d_double_array(f,nmax,3,"atom:f");

  molecule = (int *) 
    memory->srealloc(molecule,nmax*sizeof(int),"atom:molecule");

  num_bond = (int *) 
    memory->srealloc(num_bond,nmax*sizeof(int),"atom:num_bond");
  bond_type = 
    memory->grow_2d_int_array(bond_type,nmax,bond_per_atom,"atom:bond_type");
  bond_atom = 
    memory->grow_2d_int_array(bond_atom,nmax,bond_per_atom,"atom:bond_atom");

  num_angle = (int *) 
    memory->srealloc(num_angle,nmax*sizeof(int),"atom:num_angle");
  angle_type = memory->grow_2d_int_array(angle_type,nmax,angle_per_atom,
					 "atom:angle_type");
  angle_atom1 = memory->grow_2d_int_array(angle_atom1,nmax,angle_per_atom,
					  "atom:angle_atom1");
  angle_atom2 = memory->grow_2d_int_array(angle_atom2,nmax,angle_per_atom,
					  "atom:angle_atom2");
  angle_atom3 = memory->grow_2d_int_array(angle_atom3,nmax,angle_per_atom,
					  "atom:angle_atom3");

  nspecial = memory->grow_2d_int_array(nspecial,nmax,3,"atom:nspecial");
  special = memory->grow_2d_int_array(special,nmax,maxspecial,"atom:special");

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      modify->fix[extra_grow[iextra]]->grow_arrays(nmax);
}

/* ----------------------------------------------------------------------
   return # of bytes of allocated memory 
------------------------------------------------------------------------- */

int AtomAngle::memory_usage()
{
  int bytes = 0;

  bytes += natoms * sizeof(int);
  bytes += 4 * nmax * sizeof(int);
  bytes += 3 * nmax*3 * sizeof(double);
  bytes += 3 * nmax * sizeof(int);
  bytes += 2 * nmax*bond_per_atom * sizeof(int);
  bytes += 4 * nmax*angle_per_atom * sizeof(int);
  bytes += nmax*3 * sizeof(int);
  bytes += nmax*maxspecial * sizeof(int);

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      bytes += modify->fix[extra_grow[iextra]]->memory_usage(nmax);

  return bytes;
}

/* ---------------------------------------------------------------------- */

void AtomAngle::copy(int i, int j)
{
  int k;

  tag[j] = tag[i];
  type[j] = type[i];
  mask[j] = mask[i];
  image[j] = image[i];
  x[j][0] = x[i][0];
  x[j][1] = x[i][1];
  x[j][2] = x[i][2];
  v[j][0] = v[i][0];
  v[j][1] = v[i][1];
  v[j][2] = v[i][2];

  molecule[j] = molecule[i];
  num_bond[j] = num_bond[i];
  num_angle[j] = num_angle[i];
  nspecial[j][0] = nspecial[i][0];
  nspecial[j][1] = nspecial[i][1];
  nspecial[j][2] = nspecial[i][2];

  for (k = 0; k < num_bond[j]; k++) {
    bond_type[j][k] = bond_type[i][k];
    bond_atom[j][k] = bond_atom[i][k];
  }

  for (k = 0; k < num_angle[j]; k++) {
    angle_type[j][k] = angle_type[i][k];
    angle_atom1[j][k] = angle_atom1[i][k];
    angle_atom2[j][k] = angle_atom2[i][k];
    angle_atom3[j][k] = angle_atom3[i][k];
  }

  for (k = 0; k < nspecial[j][2]; k++) special[j][k] = special[i][k];

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      modify->fix[extra_grow[iextra]]->copy_arrays(i,j);
}

/* ---------------------------------------------------------------------- */

void AtomAngle::pack_comm(int n, int *list, double *buf, int *pbc_flags)
{
  int i,j,m;

  m = 0;
  if (pbc_flags[0] == 0) {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0];
      buf[m++] = x[j][1];
      buf[m++] = x[j][2];
    }
  } else {
    double xprd = domain->xprd;
    double yprd = domain->yprd;
    double zprd = domain->zprd;
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0] + pbc_flags[1]*xprd;
      buf[m++] = x[j][1] + pbc_flags[2]*yprd;
      buf[m++] = x[j][2] + pbc_flags[3]*zprd;
    }
  }
}

/* ---------------------------------------------------------------------- */

void AtomAngle::unpack_comm(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    x[i][0] = buf[m++];
    x[i][1] = buf[m++];
    x[i][2] = buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

void AtomAngle::pack_reverse(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    buf[m++] = f[i][0];
    buf[m++] = f[i][1];
    buf[m++] = f[i][2];
  }
}

/* ---------------------------------------------------------------------- */

void AtomAngle::unpack_reverse(int n, int *list, double *buf)
{
  int i,j,m;

  m = 0;
  for (i = 0; i < n; i++) {
    j = list[i];
    f[j][0] += buf[m++];
    f[j][1] += buf[m++];
    f[j][2] += buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

int AtomAngle::pack_border(int i, double *buf, int *pbc_flags)
{
  int m = 0;
  if (pbc_flags[0] == 0) {
    buf[m++] = x[i][0];
    buf[m++] = x[i][1];
    buf[m++] = x[i][2];
  } else {
    buf[m++] = x[i][0] + pbc_flags[1]*domain->xprd;
    buf[m++] = x[i][1] + pbc_flags[2]*domain->yprd;
    buf[m++] = x[i][2] + pbc_flags[3]*domain->zprd;
  }
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (molecule[i]);
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomAngle::unpack_border(int i, double *buf)
{
  if (i == nmax) grow(0);

  int m = 0;
  x[i][0] = buf[m++];
  x[i][1] = buf[m++];
  x[i][2] = buf[m++];
  tag[i] = static_cast<int> (buf[m++]);
  type[i] = static_cast<int> (buf[m++]);
  mask[i] = static_cast<int> (buf[m++]);
  molecule[i] = static_cast<int> (buf[m++]);
  return m;
}

/* ----------------------------------------------------------------------
   pack all atom quantities for shipping to another proc
   xyz must be 1st 3 values, so that comm::exchange can test on them
------------------------------------------------------------------------- */

int AtomAngle::pack_exchange(int i, double *buf)
{
  int k;

  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (image[i]);
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];

  buf[m++] = static_cast<double> (molecule[i]);

  buf[m++] = static_cast<double> (num_bond[i]);
  for (k = 0; k < num_bond[i]; k++) {
    buf[m++] = bond_type[i][k];
    buf[m++] = bond_atom[i][k];
  }

  buf[m++] = static_cast<double> (num_angle[i]);
  for (k = 0; k < num_angle[i]; k++) {
    buf[m++] = angle_type[i][k];
    buf[m++] = angle_atom1[i][k];
    buf[m++] = angle_atom2[i][k];
    buf[m++] = angle_atom3[i][k];
  }

  buf[m++] = static_cast<double> (nspecial[i][0]);
  buf[m++] = static_cast<double> (nspecial[i][1]);
  buf[m++] = static_cast<double> (nspecial[i][2]);
  for (k = 0; k < nspecial[i][2]; k++) buf[m++] = special[i][k];

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      m += modify->fix[extra_grow[iextra]]->pack_exchange(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomAngle::unpack_exchange(double *buf)
{
  int k;
  if (nlocal == nmax) grow(0);

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];

  molecule[nlocal] = static_cast<int> (buf[m++]);

  num_bond[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_bond[nlocal]; k++) {
    bond_type[nlocal][k] = static_cast<int> (buf[m++]);
    bond_atom[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_angle[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_angle[nlocal]; k++) {
    angle_type[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom3[nlocal][k] = static_cast<int> (buf[m++]);
  }

  nspecial[nlocal][0] = static_cast<int> (buf[m++]);
  nspecial[nlocal][1] = static_cast<int> (buf[m++]);
  nspecial[nlocal][2] = static_cast<int> (buf[m++]);
  for (k = 0; k < nspecial[nlocal][2]; k++) 
    special[nlocal][k] = static_cast<int> (buf[m++]);

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      m += modify->fix[extra_grow[iextra]]->unpack_exchange(nlocal,&buf[m]);

  nlocal++;
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomAngle::size_restart()
{
  int i;
  int n = 0;
  for (i = 0; i < nlocal; i++) n += 14 + 2*num_bond[i] + 4*num_angle[i];

  if (nextra_restart)
    for (int iextra = 0; iextra < nextra_restart; iextra++) 
      for (i = 0; i < nlocal; i++)
	n += modify->fix[extra_restart[iextra]]->size_restart(i);

  return n;
}

/* ----------------------------------------------------------------------
   pack all atom quantities for restart file including extra quantities
   xyz must be 1st 3 values, so that read_restart can test on them
   bond/angle types may be negative, but write as positive 
------------------------------------------------------------------------- */

int AtomAngle::pack_restart(int i, double *buf)
{
  int k;

  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (image[i]);
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];

  buf[m++] = static_cast<double> (molecule[i]);

  buf[m++] = static_cast<double> (num_bond[i]);
  for (k = 0; k < num_bond[i]; k++) {
    buf[m++] = MAX(bond_type[i][k],-bond_type[i][k]);
    buf[m++] = bond_atom[i][k];
  }

  buf[m++] = static_cast<double> (num_angle[i]);
  for (k = 0; k < num_angle[i]; k++) {
    buf[m++] = MAX(angle_type[i][k],-angle_type[i][k]);
    buf[m++] = angle_atom1[i][k];
    buf[m++] = angle_atom2[i][k];
    buf[m++] = angle_atom3[i][k];
  }

  if (nextra_restart)
    for (int iextra = 0; iextra < nextra_restart; iextra++) 
      m += modify->fix[extra_restart[iextra]]->pack_restart(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ----------------------------------------------------------------------
   unpack all atom quantities from restart file including extra quantities 
------------------------------------------------------------------------- */

int AtomAngle::unpack_restart(double *buf)
{
  int k;
  if (nlocal == nmax) {
    grow(0);
    if (nextra_store)
      extra = memory->grow_2d_double_array(extra,nmax,nextra_store,
					   "atom:extra");
  }

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];

  molecule[nlocal] = static_cast<int> (buf[m++]);

  num_bond[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_bond[nlocal]; k++) {
    bond_type[nlocal][k] = static_cast<int> (buf[m++]);
    bond_atom[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_angle[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_angle[nlocal]; k++) {
    angle_type[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom3[nlocal][k] = static_cast<int> (buf[m++]);
  }

  if (nextra_store) {
    int size = static_cast<int> (buf[0]) - m;
    for (int i = 0; i < size; i++) extra[nlocal][i] = buf[m++];
  }

  nlocal++;
  return m;
}

/* ----------------------------------------------------------------------
   unpack n lines of atoms from buf 
------------------------------------------------------------------------- */

void AtomAngle::unpack_data(int n, char *buf)
{
  int tagtmp,moltmp,typetmp,imagetmp;
  double xtmp,ytmp,ztmp;
  int imagex,imagey,imagez;
  char *next;

  double subxlo = domain->subxlo;
  double subxhi = domain->subxhi;
  double subylo = domain->subylo;
  double subyhi = domain->subyhi;
  double subzlo = domain->subzlo;
  double subzhi = domain->subzhi;

  for (int i = 0; i < n; i++) {

    next = strchr(buf,'\n');
    *next = '\0';

    if (size_atom_actual == 6) {
      sscanf(buf,"%d %d %d %lg %lg %lg",&tagtmp,&moltmp,&typetmp,
	     &xtmp,&ytmp,&ztmp);
      imagetmp = 500500500;
    } else {
      sscanf(buf,"%d %d %d %lg %lg %lg %d %d %d",
	     &tagtmp,&moltmp,&typetmp,
	     &xtmp,&ytmp,&ztmp,&imagex,&imagey,&imagez);
      imagetmp = 1000000 * (500+imagez) + 1000 * (500+imagey) + 500+imagex;
    }

    domain->remap(xtmp,ytmp,ztmp,imagetmp);
    if (xtmp >= subxlo && xtmp < subxhi &&
	ytmp >= subylo && ytmp < subyhi &&
	ztmp >= subzlo && ztmp < subzhi) {
      if (nlocal == nmax) grow(0);
      tag[nlocal] = tagtmp;
      molecule[nlocal] = moltmp;
      type[nlocal] = typetmp;
      x[nlocal][0] = xtmp;
      x[nlocal][1] = ytmp;
      x[nlocal][2] = ztmp;
      mask[nlocal] = 1;
      image[nlocal] = imagetmp;
      num_bond[nlocal] = 0;
      num_angle[nlocal] = 0;
      v[nlocal][0] = 0.0;
      v[nlocal][1] = 0.0;
      v[nlocal][2] = 0.0;
      nlocal++;
    }

    buf = next + 1;
  }
}

/* ----------------------------------------------------------------------
   create one atom of type itype at px,py,pz 
------------------------------------------------------------------------- */

void AtomAngle::create_one(int itype, double px, double py, double pz)
{
  if (nlocal == nmax) grow(0);
  tag[nlocal] = 0;
  molecule[nlocal] = 0;
  type[nlocal] = itype;
  x[nlocal][0] = px;
  x[nlocal][1] = py;
  x[nlocal][2] = pz;
  mask[nlocal] = 1;
  image[nlocal] = 500500500;
  num_bond[nlocal] = 0;
  num_angle[nlocal] = 0;
  v[nlocal][0] = 0.0;
  v[nlocal][1] = 0.0;
  v[nlocal][2] = 0.0;
  nlocal++;
}
