/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "mpi.h"
#include "atom_full.h"
#include "domain.h"
#include "force.h"
#include "modify.h"
#include "fix.h"
#include "output.h"
#include "error.h"
#include "memory.h"
#include "comm.h"

#define DELTA 10000
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

/* ---------------------------------------------------------------------- */

AtomFull::AtomFull(char *style_in) : Atom(style_in)
{
  molecular = 1;
  bonds_allow = angles_allow = dihedrals_allow = impropers_allow = 1;
  mass_require = 1;
  dipole_require = 0;

  size_comm = 3;
  size_reverse = 3;
  size_border = 8;
  size_atom_valid = 7;

  tag = type = mask = image = NULL;
  x = v = f = NULL;

  q = NULL;

  molecule = NULL;

  num_bond = NULL;
  bond_type = bond_atom = NULL;

  num_angle = NULL;
  angle_type = angle_atom1 = angle_atom2 = angle_atom3 = NULL;

  num_dihedral = NULL;
  dihedral_type = dihedral_atom1 = dihedral_atom2 = NULL;
  dihedral_atom3 = dihedral_atom4 = NULL;

  num_improper = NULL;
  improper_type = improper_atom1 = improper_atom2 = NULL;
  improper_atom3 = improper_atom4 = NULL;

  maxspecial = 1;
  nspecial = NULL;
  special = NULL;
}

/* ---------------------------------------------------------------------- */

AtomFull::~AtomFull()
{
  memory->sfree(tag);
  memory->sfree(type);
  memory->sfree(mask);
  memory->sfree(image);
  memory->destroy_2d_double_array(x);
  memory->destroy_2d_double_array(v);
  memory->destroy_2d_double_array(f);

  memory->sfree(q);

  memory->sfree(molecule);

  memory->sfree(num_bond);
  memory->destroy_2d_int_array(bond_type);
  memory->destroy_2d_int_array(bond_atom);

  memory->sfree(num_angle);
  memory->destroy_2d_int_array(angle_type);
  memory->destroy_2d_int_array(angle_atom1);
  memory->destroy_2d_int_array(angle_atom2);
  memory->destroy_2d_int_array(angle_atom3);

  memory->sfree(num_dihedral);
  memory->destroy_2d_int_array(dihedral_type);
  memory->destroy_2d_int_array(dihedral_atom1);
  memory->destroy_2d_int_array(dihedral_atom2);
  memory->destroy_2d_int_array(dihedral_atom3);
  memory->destroy_2d_int_array(dihedral_atom4);

  memory->sfree(num_improper);
  memory->destroy_2d_int_array(improper_type);
  memory->destroy_2d_int_array(improper_atom1);
  memory->destroy_2d_int_array(improper_atom2);
  memory->destroy_2d_int_array(improper_atom3);
  memory->destroy_2d_int_array(improper_atom4);

  memory->destroy_2d_int_array(nspecial);
  memory->destroy_2d_int_array(special);
}

/* ----------------------------------------------------------------------
   grow atom arrays
   n = 0 grows arrays by DELTA
   n > 0 allocates arrays to size n  
------------------------------------------------------------------------- */

void AtomFull::grow(int n)
{
  if (n == 0) nmax += DELTA;
  else nmax = n;

  tag = (int *) memory->srealloc(tag,nmax*sizeof(int),"atom:tag");
  type = (int *) memory->srealloc(type,nmax*sizeof(int),"atom:type");
  mask = (int *) memory->srealloc(mask,nmax*sizeof(int),"atom:mask");
  image = (int *) memory->srealloc(image,nmax*sizeof(int),"atom:image");
  x = memory->grow_2d_double_array(x,nmax,3,"atom:x");
  v = memory->grow_2d_double_array(v,nmax,3,"atom:v");
  f = memory->grow_2d_double_array(f,nmax,3,"atom:f");

  q = (double *) memory->srealloc(q,nmax*sizeof(double),"atom:q");

  molecule = (int *) 
    memory->srealloc(molecule,nmax*sizeof(int),"atom:molecule");

  num_bond = (int *) 
    memory->srealloc(num_bond,nmax*sizeof(int),"atom:num_bond");
  bond_type = 
    memory->grow_2d_int_array(bond_type,nmax,bond_per_atom,"atom:bond_type");
  bond_atom = 
    memory->grow_2d_int_array(bond_atom,nmax,bond_per_atom,"atom:bond_atom");

  num_angle = (int *) 
    memory->srealloc(num_angle,nmax*sizeof(int),"atom:num_angle");
  angle_type = memory->grow_2d_int_array(angle_type,nmax,angle_per_atom,
					 "atom:angle_type");
  angle_atom1 = memory->grow_2d_int_array(angle_atom1,nmax,angle_per_atom,
					  "atom:angle_atom1");
  angle_atom2 = memory->grow_2d_int_array(angle_atom2,nmax,angle_per_atom,
					  "atom:angle_atom2");
  angle_atom3 = memory->grow_2d_int_array(angle_atom3,nmax,angle_per_atom,
					  "atom:angle_atom3");

  num_dihedral = (int *) 
    memory->srealloc(num_dihedral,nmax*sizeof(int),"atom:num_dihedral");
  dihedral_type = 
    memory->grow_2d_int_array(dihedral_type,nmax,dihedral_per_atom,
			      "atom:dihedral_type");
  dihedral_atom1 = 
    memory->grow_2d_int_array(dihedral_atom1,nmax,dihedral_per_atom,
			      "atom:dihedral_atom1");
  dihedral_atom2 = 
    memory->grow_2d_int_array(dihedral_atom2,nmax,dihedral_per_atom,
			      "atom:dihedral_atom2");
  dihedral_atom3 = 
    memory->grow_2d_int_array(dihedral_atom3,nmax,dihedral_per_atom,
			      "atom:dihedral_atom3");
  dihedral_atom4 = 
    memory->grow_2d_int_array(dihedral_atom4,nmax,dihedral_per_atom,
			      "atom:dihedral_atom4");

  num_improper = (int *) 
    memory->srealloc(num_improper,nmax*sizeof(int),"atom:num_improper");
  improper_type = 
    memory->grow_2d_int_array(improper_type,nmax,improper_per_atom,
			      "atom:improper_type");
  improper_atom1 = 
    memory->grow_2d_int_array(improper_atom1,nmax,improper_per_atom,
			      "atom:improper_atom1");
  improper_atom2 = 
    memory->grow_2d_int_array(improper_atom2,nmax,improper_per_atom,
			      "atom:improper_atom2");
  improper_atom3 = 
    memory->grow_2d_int_array(improper_atom3,nmax,improper_per_atom,
			      "atom:improper_atom3");
  improper_atom4 = 
    memory->grow_2d_int_array(improper_atom4,nmax,improper_per_atom,
			      "atom:improper_atom4");

  nspecial = memory->grow_2d_int_array(nspecial,nmax,3,"atom:nspecial");
  special = memory->grow_2d_int_array(special,nmax,maxspecial,"atom:special");

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      modify->fix[extra_grow[iextra]]->grow_arrays(nmax);
}

/* ----------------------------------------------------------------------
   return # of bytes of allocated memory 
------------------------------------------------------------------------- */

int AtomFull::memory_usage()
{
  int bytes = 0;

  bytes += natoms * sizeof(int);
  bytes += 4 * nmax * sizeof(int);
  bytes += 3 * nmax*3 * sizeof(double);
  bytes += 1 * nmax * sizeof(double);
  bytes += 5 * nmax * sizeof(int);
  bytes += 2 * nmax*bond_per_atom * sizeof(int);
  bytes += 4 * nmax*angle_per_atom * sizeof(int);
  bytes += 5 * nmax*dihedral_per_atom * sizeof(int);
  bytes += 5 * nmax*improper_per_atom * sizeof(int);
  bytes += nmax*3 * sizeof(int);
  bytes += nmax*maxspecial * sizeof(int);

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++)
      bytes += modify->fix[extra_grow[iextra]]->memory_usage(nmax);

  return bytes;
}

/* ---------------------------------------------------------------------- */

void AtomFull::copy(int i, int j)
{
  int k;

  tag[j] = tag[i];
  type[j] = type[i];
  mask[j] = mask[i];
  image[j] = image[i];
  x[j][0] = x[i][0];
  x[j][1] = x[i][1];
  x[j][2] = x[i][2];
  v[j][0] = v[i][0];
  v[j][1] = v[i][1];
  v[j][2] = v[i][2];

  q[j] = q[i];

  molecule[j] = molecule[i];
  num_bond[j] = num_bond[i];
  num_angle[j] = num_angle[i];
  num_dihedral[j] = num_dihedral[i];
  num_improper[j] = num_improper[i];
  nspecial[j][0] = nspecial[i][0];
  nspecial[j][1] = nspecial[i][1];
  nspecial[j][2] = nspecial[i][2];

  for (k = 0; k < num_bond[j]; k++) {
    bond_type[j][k] = bond_type[i][k];
    bond_atom[j][k] = bond_atom[i][k];
  }

  for (k = 0; k < num_angle[j]; k++) {
    angle_type[j][k] = angle_type[i][k];
    angle_atom1[j][k] = angle_atom1[i][k];
    angle_atom2[j][k] = angle_atom2[i][k];
    angle_atom3[j][k] = angle_atom3[i][k];
  }

  for (k = 0; k < num_dihedral[j]; k++) {
    dihedral_type[j][k] = dihedral_type[i][k];
    dihedral_atom1[j][k] = dihedral_atom1[i][k];
    dihedral_atom2[j][k] = dihedral_atom2[i][k];
    dihedral_atom3[j][k] = dihedral_atom3[i][k];
    dihedral_atom4[j][k] = dihedral_atom4[i][k];
  }

  for (k = 0; k < num_improper[j]; k++) {
    improper_type[j][k] = improper_type[i][k];
    improper_atom1[j][k] = improper_atom1[i][k];
    improper_atom2[j][k] = improper_atom2[i][k];
    improper_atom3[j][k] = improper_atom3[i][k];
    improper_atom4[j][k] = improper_atom4[i][k];
  }

  for (k = 0; k < nspecial[j][2]; k++) special[j][k] = special[i][k];

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      modify->fix[extra_grow[iextra]]->copy_arrays(i,j);
}

/* ---------------------------------------------------------------------- */

void AtomFull::pack_comm(int n, int *list, double *buf, int *pbc_flags)
{
  int i,j,m;

  m = 0;
  if (pbc_flags[0] == 0) {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0];
      buf[m++] = x[j][1];
      buf[m++] = x[j][2];
    }
  } else {
    double xprd = domain->xprd;
    double yprd = domain->yprd;
    double zprd = domain->zprd;
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0] + pbc_flags[1]*xprd;
      buf[m++] = x[j][1] + pbc_flags[2]*yprd;
      buf[m++] = x[j][2] + pbc_flags[3]*zprd;
    }
  }
}

/* ---------------------------------------------------------------------- */

void AtomFull::unpack_comm(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    x[i][0] = buf[m++];
    x[i][1] = buf[m++];
    x[i][2] = buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

void AtomFull::pack_reverse(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    buf[m++] = f[i][0];
    buf[m++] = f[i][1];
    buf[m++] = f[i][2];
  }
}

/* ---------------------------------------------------------------------- */

void AtomFull::unpack_reverse(int n, int *list, double *buf)
{
  int i,j,m;

  m = 0;
  for (i = 0; i < n; i++) {
    j = list[i];
    f[j][0] += buf[m++];
    f[j][1] += buf[m++];
    f[j][2] += buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

int AtomFull::pack_border(int i, double *buf, int *pbc_flags)
{
  int m = 0;
  if (pbc_flags[0] == 0) {
    buf[m++] = x[i][0];
    buf[m++] = x[i][1];
    buf[m++] = x[i][2];
  } else {
    buf[m++] = x[i][0] + pbc_flags[1]*domain->xprd;
    buf[m++] = x[i][1] + pbc_flags[2]*domain->yprd;
    buf[m++] = x[i][2] + pbc_flags[3]*domain->zprd;
  }
  buf[m++] = q[i];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (molecule[i]);
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomFull::unpack_border(int i, double *buf)
{
  if (i == nmax) grow(0);

  int m = 0;
  x[i][0] = buf[m++];
  x[i][1] = buf[m++];
  x[i][2] = buf[m++];
  q[i] = buf[m++];
  tag[i] = static_cast<int> (buf[m++]);
  type[i] = static_cast<int> (buf[m++]);
  mask[i] = static_cast<int> (buf[m++]);
  molecule[i] = static_cast<int> (buf[m++]);
  return m;
}

/* ----------------------------------------------------------------------
   pack all atom quantities for shipping to another proc
   xyz must be 1st 3 values, so that comm::exchange can test on them
------------------------------------------------------------------------- */

int AtomFull::pack_exchange(int i, double *buf)
{
  int k;

  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (image[i]);
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];

  buf[m++] = q[i];

  buf[m++] = static_cast<double> (molecule[i]);

  buf[m++] = static_cast<double> (num_bond[i]);
  for (k = 0; k < num_bond[i]; k++) {
    buf[m++] = bond_type[i][k];
    buf[m++] = bond_atom[i][k];
  }

  buf[m++] = static_cast<double> (num_angle[i]);
  for (k = 0; k < num_angle[i]; k++) {
    buf[m++] = angle_type[i][k];
    buf[m++] = angle_atom1[i][k];
    buf[m++] = angle_atom2[i][k];
    buf[m++] = angle_atom3[i][k];
  }

  buf[m++] = static_cast<double> (num_dihedral[i]);
  for (k = 0; k < num_dihedral[i]; k++) {
    buf[m++] = dihedral_type[i][k];
    buf[m++] = dihedral_atom1[i][k];
    buf[m++] = dihedral_atom2[i][k];
    buf[m++] = dihedral_atom3[i][k];
    buf[m++] = dihedral_atom4[i][k];
  }

  buf[m++] = static_cast<double> (num_improper[i]);
  for (k = 0; k < num_improper[i]; k++) {
    buf[m++] = improper_type[i][k];
    buf[m++] = improper_atom1[i][k];
    buf[m++] = improper_atom2[i][k];
    buf[m++] = improper_atom3[i][k];
    buf[m++] = improper_atom4[i][k];
  }

  buf[m++] = static_cast<double> (nspecial[i][0]);
  buf[m++] = static_cast<double> (nspecial[i][1]);
  buf[m++] = static_cast<double> (nspecial[i][2]);
  for (k = 0; k < nspecial[i][2]; k++) buf[m++] = special[i][k];

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      m += modify->fix[extra_grow[iextra]]->pack_exchange(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomFull::unpack_exchange(double *buf)
{
  int k;
  if (nlocal == nmax) grow(0);

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];

  q[nlocal] = buf[m++];

  molecule[nlocal] = static_cast<int> (buf[m++]);

  num_bond[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_bond[nlocal]; k++) {
    bond_type[nlocal][k] = static_cast<int> (buf[m++]);
    bond_atom[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_angle[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_angle[nlocal]; k++) {
    angle_type[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom3[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_dihedral[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_dihedral[nlocal]; k++) {
    dihedral_type[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom3[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom4[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_improper[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_improper[nlocal]; k++) {
    improper_type[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom3[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom4[nlocal][k] = static_cast<int> (buf[m++]);
  }

  nspecial[nlocal][0] = static_cast<int> (buf[m++]);
  nspecial[nlocal][1] = static_cast<int> (buf[m++]);
  nspecial[nlocal][2] = static_cast<int> (buf[m++]);
  for (k = 0; k < nspecial[nlocal][2]; k++) 
    special[nlocal][k] = static_cast<int> (buf[m++]);

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      m += modify->fix[extra_grow[iextra]]->unpack_exchange(nlocal,&buf[m]);

  nlocal++;
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomFull::size_restart()
{
  int i;

  int n = 0;
  for (i = 0; i < nlocal; i++) 
    n += 17 + 2*num_bond[i] + 4*num_angle[i] + 
      5*num_dihedral[i] + 5*num_improper[i];

  if (nextra_restart)
    for (int iextra = 0; iextra < nextra_restart; iextra++) 
      for (i = 0; i < nlocal; i++)
	n += modify->fix[extra_restart[iextra]]->size_restart(i);

  return n;
}

/* ----------------------------------------------------------------------
   pack all atom quantities for restart file including extra quantities
   xyz must be 1st 3 values, so that read_restart can test on them
   interaction types may be negative, but write as positive   
------------------------------------------------------------------------- */

int AtomFull::pack_restart(int i, double *buf)
{
  int k;

  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (image[i]);
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];

  buf[m++] = q[i];

  buf[m++] = static_cast<double> (molecule[i]);

  buf[m++] = static_cast<double> (num_bond[i]);
  for (k = 0; k < num_bond[i]; k++) {
    buf[m++] = MAX(bond_type[i][k],-bond_type[i][k]);
    buf[m++] = bond_atom[i][k];
  }

  buf[m++] = static_cast<double> (num_angle[i]);
  for (k = 0; k < num_angle[i]; k++) {
    buf[m++] = MAX(angle_type[i][k],-angle_type[i][k]);
    buf[m++] = angle_atom1[i][k];
    buf[m++] = angle_atom2[i][k];
    buf[m++] = angle_atom3[i][k];
  }

  buf[m++] = static_cast<double> (num_dihedral[i]);
  for (k = 0; k < num_dihedral[i]; k++) {
    buf[m++] = MAX(dihedral_type[i][k],-dihedral_type[i][k]);
    buf[m++] = dihedral_atom1[i][k];
    buf[m++] = dihedral_atom2[i][k];
    buf[m++] = dihedral_atom3[i][k];
    buf[m++] = dihedral_atom4[i][k];
  }

  buf[m++] = static_cast<double> (num_improper[i]);
  for (k = 0; k < num_improper[i]; k++) {
    buf[m++] = MAX(improper_type[i][k],-improper_type[i][k]);
    buf[m++] = improper_atom1[i][k];
    buf[m++] = improper_atom2[i][k];
    buf[m++] = improper_atom3[i][k];
    buf[m++] = improper_atom4[i][k];
  }

  if (nextra_restart)
    for (int iextra = 0; iextra < nextra_restart; iextra++) 
      m += modify->fix[extra_restart[iextra]]->pack_restart(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ----------------------------------------------------------------------
   unpack all atom quantities from restart file including extra quantities
------------------------------------------------------------------------- */

int AtomFull::unpack_restart(double *buf)
{
  int k;
  if (nlocal == nmax) {
    grow(0);
    if (nextra_store)
      extra = memory->grow_2d_double_array(extra,nmax,nextra_store,
					   "atom:extra");
  }

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];

  q[nlocal] = buf[m++];

  molecule[nlocal] = static_cast<int> (buf[m++]);

  num_bond[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_bond[nlocal]; k++) {
    bond_type[nlocal][k] = static_cast<int> (buf[m++]);
    bond_atom[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_angle[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_angle[nlocal]; k++) {
    angle_type[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    angle_atom3[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_dihedral[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_dihedral[nlocal]; k++) {
    dihedral_type[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom3[nlocal][k] = static_cast<int> (buf[m++]);
    dihedral_atom4[nlocal][k] = static_cast<int> (buf[m++]);
  }

  num_improper[nlocal] = static_cast<int> (buf[m++]);
  for (k = 0; k < num_improper[nlocal]; k++) {
    improper_type[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom1[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom2[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom3[nlocal][k] = static_cast<int> (buf[m++]);
    improper_atom4[nlocal][k] = static_cast<int> (buf[m++]);
  }

  if (nextra_store) {
    int size = static_cast<int> (buf[0]) - m;
    for (int i = 0; i < size; i++) extra[nlocal][i] = buf[m++];
  }

  nlocal++;
  return m;
}

/* ----------------------------------------------------------------------
   unpack n lines of atoms from buf 
------------------------------------------------------------------------- */

void AtomFull::unpack_data(int n, char *buf)
{
  int tagtmp,moltmp,typetmp,imagetmp;
  double qtmp,xtmp,ytmp,ztmp;
  int imagex,imagey,imagez;
  char *next;

  double subxlo = domain->subxlo;
  double subxhi = domain->subxhi;
  double subylo = domain->subylo;
  double subyhi = domain->subyhi;
  double subzlo = domain->subzlo;
  double subzhi = domain->subzhi;

  for (int i = 0; i < n; i++) {

    next = strchr(buf,'\n');
    *next = '\0';

    if (size_atom_actual == 7) {
      sscanf(buf,"%d %d %d %lg %lg %lg %lg",&tagtmp,&moltmp,&typetmp,
	     &qtmp,&xtmp,&ytmp,&ztmp);
      imagetmp = 500500500;
    } else {
      sscanf(buf,"%d %d %d %lg %lg %lg %lg %d %d %d",
	     &tagtmp,&moltmp,&typetmp,
	     &qtmp,&xtmp,&ytmp,&ztmp,&imagex,&imagey,&imagez);
      imagetmp = 1000000 * (500+imagez) + 1000 * (500+imagey) + 500+imagex;
    }

    domain->remap(xtmp,ytmp,ztmp,imagetmp);
    if (xtmp >= subxlo && xtmp < subxhi &&
	ytmp >= subylo && ytmp < subyhi &&
	ztmp >= subzlo && ztmp < subzhi) {
      if (nlocal == nmax) grow(0);
      tag[nlocal] = tagtmp;
      molecule[nlocal] = moltmp;
      type[nlocal] = typetmp;
      q[nlocal] = qtmp;
      x[nlocal][0] = xtmp;
      x[nlocal][1] = ytmp;
      x[nlocal][2] = ztmp;
      mask[nlocal] = 1;
      image[nlocal] = imagetmp;
      num_bond[nlocal] = 0;
      num_angle[nlocal] = 0;
      num_dihedral[nlocal] = 0;
      num_improper[nlocal] = 0;
      v[nlocal][0] = 0.0;
      v[nlocal][1] = 0.0;
      v[nlocal][2] = 0.0;
      nlocal++;
    }

    buf = next + 1;
  }
}

/* ----------------------------------------------------------------------
   create one atom of type itype at px,py,pz
------------------------------------------------------------------------- */

void AtomFull::create_one(int itype, double px, double py, double pz)
{
  if (nlocal == nmax) grow(0);
  tag[nlocal] = 0;
  molecule[nlocal] = 0;
  type[nlocal] = itype;
  q[nlocal] = 0.0;
  x[nlocal][0] = px;
  x[nlocal][1] = py;
  x[nlocal][2] = pz;
  mask[nlocal] = 1;
  image[nlocal] = 500500500;
  num_bond[nlocal] = 0;
  num_angle[nlocal] = 0;
  num_dihedral[nlocal] = 0;
  num_improper[nlocal] = 0;
  v[nlocal][0] = 0.0;
  v[nlocal][1] = 0.0;
  v[nlocal][2] = 0.0;
  nlocal++;
}
