/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "mpi.h"
#include "atom.h"
#include "comm.h"
#include "force.h"
#include "modify.h"
#include "fix.h"
#include "output.h"
#include "domain.h"
#include "group.h"
#include "memory.h"
#include "error.h"

#define DELTA_CALLBACK 1

#define MIN(A,B) ((A) < (B)) ? (A) : (B)
#define MAX(A,B) ((A) > (B)) ? (A) : (B)

/* ---------------------------------------------------------------------- */

Atom::Atom(char *style_in)
{
  int n = strlen(style_in) + 1;
  style = new char[n];
  strcpy(style,style_in);

  // set data and restart file header values to defaults

  natoms = 0;
  nbonds = nangles = ndihedrals = nimpropers = 0;
  ntypes = 0;
  nbondtypes = nangletypes = ndihedraltypes = nimpropertypes = 0;
  bond_per_atom = angle_per_atom = dihedral_per_atom = improper_per_atom = 0;

  // no atoms initially

  nlocal = 0;
  nghost = 0;
  nmax = 0;
  map = NULL;

  // arrays of length ntypes

  mass = NULL;
  mass_flag = NULL;
  dipole = NULL;
  dipole_flag = NULL;

  // callback lists & extra restart info

  nextra_grow = nextra_restart = 0;
  extra_grow = extra_restart = NULL;
  nextra_grow_max = nextra_restart_max = 0;
  nextra_store = 0;
  extra = NULL;
}

/* ---------------------------------------------------------------------- */

Atom::~Atom()
{
  delete [] style;
  memory->sfree(map);

  delete [] mass;
  delete [] mass_flag;
  delete [] dipole;
  delete [] dipole_flag;

  memory->sfree(extra_grow);
  memory->sfree(extra_restart);
  memory->destroy_2d_double_array(extra);
}

/* ---------------------------------------------------------------------- */

void Atom::init()
{
  if (mass_require) check_mass();
  if (dipole_require) check_dipole();

  // delete extra array since it doesn't persist past first run

  if (nextra_store) {
    memory->destroy_2d_double_array(extra);
    extra = NULL;
    nextra_store = 0;
  }

  init_style();
}

/* ----------------------------------------------------------------------
   allocate memory for global -> local map
   length = 1 to largest tag of any atom
   set entire array to -1 as initial value 
------------------------------------------------------------------------- */

void Atom::map_init()
{
  int max = 0;
  for (int i = 0; i < nlocal; i++) max = MAX(max,tag[i]);
  int maxall;
  MPI_Allreduce(&max,&maxall,1,MPI_INT,MPI_MAX,world);
  
  memory->sfree(map);
  map = (int *) memory->smalloc((maxall+1)*sizeof(int),"atom:map");
  for (int i = 0; i <= natoms; i++) map[i] = -1;
}

/* ----------------------------------------------------------------------
   clear out global -> local map for my own and ghost atoms 
------------------------------------------------------------------------- */

void Atom::map_clear()
{
  int nall = nlocal + nghost;
  for (int i = 0; i < nall; i++) map[tag[i]] = -1;
}

/* ----------------------------------------------------------------------
   reset global -> local map for my own and ghost atoms
   loop in reverse order so that nearby images take precedence over far ones
     and owned atoms take precedence over images
   this enables valid lookups of bond topology atoms 
------------------------------------------------------------------------- */

void Atom::map_set()
{
  int nall = nlocal + nghost;
  for (int i = nall-1; i >= 0 ; i--) map[tag[i]] = i;
}

/* ----------------------------------------------------------------------
   free the memory for the map array 
------------------------------------------------------------------------- */

void Atom::map_delete()
{
  memory->sfree(map);
  map = NULL;
}

/* ----------------------------------------------------------------------
   add unique tags to any atoms with tag = 0
   new tags are grouped by proc and start after max current tag
   called after creating new atoms 
------------------------------------------------------------------------- */

void Atom::tag_extend()
{
  // maxtag_all = max tag for all atoms

  int maxtag = 0;
  for (int i = 0; i < nlocal; i++) maxtag = MAX(maxtag,tag[i]);
  int maxtag_all;
  MPI_Allreduce(&maxtag,&maxtag_all,1,MPI_INT,MPI_MAX,world);

  // notag = # of atoms I own with no tag (tag = 0)
  // notag_sum = # of total atoms on procs <= me with no tag

  int notag = 0;
  for (int i = 0; i < nlocal; i++) if (tag[i] == 0) notag++;
  int notag_sum;
  MPI_Scan(&notag,&notag_sum,1,MPI_INT,MPI_SUM,world);

  // itag = 1st new tag that my untagged atoms should use

  int itag = maxtag_all + notag_sum - notag + 1;
  for (int i = 0; i < nlocal; i++) if (tag[i] == 0) tag[i] = itag++;
}

/* ----------------------------------------------------------------------
   parse one atom line from data file, check if valid for atom style 
------------------------------------------------------------------------- */

int Atom::parse_data(char *line)
{
  size_atom_actual = count_words(line);
  if (size_atom_actual == size_atom_valid || 
      size_atom_actual == size_atom_valid + 3) return 0;
  else return 1;
}

/* ----------------------------------------------------------------------
   count and return words in a single line
   make copy of line before using strtok so as not to change line 
------------------------------------------------------------------------- */

int Atom::count_words(char *line)
{
  int n = strlen(line) + 1;
  char *copy = (char *) memory->smalloc(n*sizeof(char),"copy");
  strcpy(copy,line);

  if (strtok(copy," \t\n") == NULL) return 0;
  n = 1;
  while (strtok(NULL," \t\n")) n++;

  memory->sfree(copy);
  return n;
}

/* ---------------------------------------------------------------------- */

void Atom::unpack_vels(int n, char *buf)
{
  int m,tagtmp;
  double vxtmp,vytmp,vztmp;
  char *next;

  for (int i = 0; i < n; i++) {
    next = strchr(buf,'\n');
    *next = '\0';
    sscanf(buf,"%d %lg %lg %lg",&tagtmp,&vxtmp,&vytmp,&vztmp);
    if ((m = map[tagtmp]) >= 0) {
      v[m][0] = vxtmp;
      v[m][1] = vytmp;
      v[m][2] = vztmp;
    }
    buf = next + 1;
  }
}

/* ---------------------------------------------------------------------- */

void Atom::unpack_bonds(int n, char *buf)
{
  int m,tmp,itype,atom1,atom2;
  char *next;
  int newton_bond = force->newton_bond;

  for (int i = 0; i < n; i++) {
    next = strchr(buf,'\n');
    *next = '\0';
    sscanf(buf,"%d %d %d %d",&tmp,&itype,&atom1,&atom2);
    if ((m = map[atom1]) >= 0) {
      bond_type[m][num_bond[m]] = itype;
      bond_atom[m][num_bond[m]] = atom2;
      num_bond[m]++;
    }
    if (newton_bond == 0) {
      if ((m = map[atom2]) >= 0) {
	bond_type[m][num_bond[m]] = itype;
	bond_atom[m][num_bond[m]] = atom1;
	num_bond[m]++;
      }
    }
    buf = next + 1;
  }
}

/* ---------------------------------------------------------------------- */

void Atom::unpack_angles(int n, char *buf)
{
  int m,tmp,itype,atom1,atom2,atom3;
  char *next;
  int newton_bond = force->newton_bond;

  for (int i = 0; i < n; i++) {
    next = strchr(buf,'\n');
    *next = '\0';
    sscanf(buf,"%d %d %d %d %d",&tmp,&itype,&atom1,&atom2,&atom3);
    if ((m = map[atom2]) >= 0) {
      angle_type[m][num_angle[m]] = itype;
      angle_atom1[m][num_angle[m]] = atom1;
      angle_atom2[m][num_angle[m]] = atom2;
      angle_atom3[m][num_angle[m]] = atom3;
      num_angle[m]++;
    }
    if (newton_bond == 0) {
      if ((m = map[atom1]) >= 0) {
	angle_type[m][num_angle[m]] = itype;
	angle_atom1[m][num_angle[m]] = atom1;
	angle_atom2[m][num_angle[m]] = atom2;
	angle_atom3[m][num_angle[m]] = atom3;
	num_angle[m]++;
      }
      if ((m = map[atom3]) >= 0) {
	angle_type[m][num_angle[m]] = itype;
	angle_atom1[m][num_angle[m]] = atom1;
	angle_atom2[m][num_angle[m]] = atom2;
	angle_atom3[m][num_angle[m]] = atom3;
	num_angle[m]++;
      }
    }
    buf = next + 1;
  }
}

/* ---------------------------------------------------------------------- */

void Atom::unpack_dihedrals(int n, char *buf)
{
  int m,tmp,itype,atom1,atom2,atom3,atom4;
  char *next;
  int newton_bond = force->newton_bond;

  for (int i = 0; i < n; i++) {
    next = strchr(buf,'\n');
    *next = '\0';
    sscanf(buf,"%d %d %d %d %d %d",&tmp,&itype,&atom1,&atom2,&atom3,&atom4);
    if ((m = map[atom2]) >= 0) {
      dihedral_type[m][num_dihedral[m]] = itype;
      dihedral_atom1[m][num_dihedral[m]] = atom1;
      dihedral_atom2[m][num_dihedral[m]] = atom2;
      dihedral_atom3[m][num_dihedral[m]] = atom3;
      dihedral_atom4[m][num_dihedral[m]] = atom4;
      num_dihedral[m]++;
    }
    if (newton_bond == 0) {
      if ((m = map[atom1]) >= 0) {
	dihedral_type[m][num_dihedral[m]] = itype;
	dihedral_atom1[m][num_dihedral[m]] = atom1;
	dihedral_atom2[m][num_dihedral[m]] = atom2;
	dihedral_atom3[m][num_dihedral[m]] = atom3;
	dihedral_atom4[m][num_dihedral[m]] = atom4;
	num_dihedral[m]++;
      }
      if ((m = map[atom3]) >= 0) {
	dihedral_type[m][num_dihedral[m]] = itype;
	dihedral_atom1[m][num_dihedral[m]] = atom1;
	dihedral_atom2[m][num_dihedral[m]] = atom2;
	dihedral_atom3[m][num_dihedral[m]] = atom3;
	dihedral_atom4[m][num_dihedral[m]] = atom4;
	num_dihedral[m]++;
      }
      if ((m = map[atom4]) >= 0) {
	dihedral_type[m][num_dihedral[m]] = itype;
	dihedral_atom1[m][num_dihedral[m]] = atom1;
	dihedral_atom2[m][num_dihedral[m]] = atom2;
	dihedral_atom3[m][num_dihedral[m]] = atom3;
	dihedral_atom4[m][num_dihedral[m]] = atom4;
	num_dihedral[m]++;
      }
    }
    buf = next + 1;
  }
}

/* ---------------------------------------------------------------------- */

void Atom::unpack_impropers(int n, char *buf)
{
  int m,tmp,itype,atom1,atom2,atom3,atom4;
  char *next;
  int newton_bond = force->newton_bond;

  for (int i = 0; i < n; i++) {
    next = strchr(buf,'\n');
    *next = '\0';
    sscanf(buf,"%d %d %d %d %d %d",&tmp,&itype,&atom1,&atom2,&atom3,&atom4);
    if ((m = map[atom2]) >= 0) {
      improper_type[m][num_improper[m]] = itype;
      improper_atom1[m][num_improper[m]] = atom1;
      improper_atom2[m][num_improper[m]] = atom2;
      improper_atom3[m][num_improper[m]] = atom3;
      improper_atom4[m][num_improper[m]] = atom4;
      num_improper[m]++;
    }
    if (newton_bond == 0) {
      if ((m = map[atom1]) >= 0) {
	improper_type[m][num_improper[m]] = itype;
	improper_atom1[m][num_improper[m]] = atom1;
	improper_atom2[m][num_improper[m]] = atom2;
	improper_atom3[m][num_improper[m]] = atom3;
	improper_atom4[m][num_improper[m]] = atom4;
	num_improper[m]++;
      }
      if ((m = map[atom3]) >= 0) {
	improper_type[m][num_improper[m]] = itype;
	improper_atom1[m][num_improper[m]] = atom1;
	improper_atom2[m][num_improper[m]] = atom2;
	improper_atom3[m][num_improper[m]] = atom3;
	improper_atom4[m][num_improper[m]] = atom4;
	num_improper[m]++;
      }
      if ((m = map[atom4]) >= 0) {
	improper_type[m][num_improper[m]] = itype;
	improper_atom1[m][num_improper[m]] = atom1;
	improper_atom2[m][num_improper[m]] = atom2;
	improper_atom3[m][num_improper[m]] = atom3;
	improper_atom4[m][num_improper[m]] = atom4;
	num_improper[m]++;
      }
    }
    buf = next + 1;
  }
}

/* ----------------------------------------------------------------------
   allocate arrays of length ntypes
   only done after ntypes is set
------------------------------------------------------------------------- */

void Atom::allocate_type_arrays()
{
  if (mass_require) {
    mass = new double[ntypes+1];
    mass_flag = new int[ntypes+1];
    for (int itype = 1; itype <= ntypes; itype++) mass_flag[itype] = 0;
  }
  if (dipole_require) {
    dipole = new double[ntypes+1];
    dipole_flag = new int[ntypes+1];
    for (int itype = 1; itype <= ntypes; itype++) dipole_flag[itype] = 0;
  }
}

/* ----------------------------------------------------------------------
   set a mass and flag it as set
   called from reading of data file
------------------------------------------------------------------------- */

void Atom::set_mass(char *str)
{
  if (mass_require == 0) error->all("Cannot set mass for this atom style");

  int i;
  double mass_one;
  sscanf(str,"%d %lg",&i,&mass_one);

  mass[i] = mass_one;
  mass_flag[i] = 1;
}

/* ----------------------------------------------------------------------
   set a mass and flag it as set
   called from EAM pair routine
------------------------------------------------------------------------- */

void Atom::set_mass(int itype, double value)
{
  if (mass_require == 0) error->all("Cannot set mass for this atom style");
  if (itype < 1 || itype > ntypes) error->all("Illegal type for mass set");

  mass[itype] = value;
  mass_flag[itype] = 1;
}

/* ----------------------------------------------------------------------
   set one or more masses and flag them as set
   called from reading of input script
------------------------------------------------------------------------- */

void Atom::set_mass(int narg, char **arg)
{
  if (mass_require == 0) error->all("Cannot set mass for this atom style");

  int lo,hi;
  force->bounds(arg[0],ntypes,lo,hi);
  if (lo < 1 || hi > ntypes) error->all("Illegal type for mass set");

  for (int itype = lo; itype <= hi; itype++) {
    mass[itype] = atof(arg[1]);
    mass_flag[itype] = 1;
  }
}

/* ----------------------------------------------------------------------
   set all masses as read in from restart file
------------------------------------------------------------------------- */

void Atom::set_mass(double *values)
{
  for (int itype = 1; itype <= ntypes; itype++) {
    mass[itype] = values[itype];
    mass_flag[itype] = 1;
  }
}

/* ----------------------------------------------------------------------
   check that all masses have been set
------------------------------------------------------------------------- */

void Atom::check_mass()
{
  for (int itype = 1; itype <= ntypes; itype++)
    if (mass_flag[itype] == 0) error->all("All masses are not set");
}

/* ----------------------------------------------------------------------
   set a dipole moment and flag it as set
   called from reading of data file
------------------------------------------------------------------------- */

void Atom::set_dipole(char *str)
{
  if (dipole_require == 0) 
    error->all("Cannot set dipole for this atom style");

  int i;
  double dipole_one;
  sscanf(str,"%d %lg",&i,&dipole_one);

  dipole[i] = dipole_one;
  dipole_flag[i] = 1;
}

/* ----------------------------------------------------------------------
   set one or more dipole moments and flag them as set
   called from reading of input script
------------------------------------------------------------------------- */

void Atom::set_dipole(int narg, char **arg)
{
  if (dipole_require == 0) 
    error->all("Cannot set dipole for this atom style");

  int lo,hi;
  force->bounds(arg[0],ntypes,lo,hi);
  if (lo < 1 || hi > ntypes) error->all("Illegal type for dipole set");

  for (int itype = lo; itype <= hi; itype++) {
    dipole[itype] = atof(arg[1]);
    dipole_flag[itype] = 1;
  }
}

/* ----------------------------------------------------------------------
   set all dipole moments as read in from restart file
------------------------------------------------------------------------- */

void Atom::set_dipole(double *values)
{
  for (int itype = 1; itype <= ntypes; itype++) {
    dipole[itype] = values[itype];
    dipole_flag[itype] = 1;
  }
}

/* ----------------------------------------------------------------------
   check that all dipole moments have been set
------------------------------------------------------------------------- */

void Atom::check_dipole()
{
  for (int itype = 1; itype <= ntypes; itype++)
    if (dipole_flag[itype] == 0) error->all("All dipole moments are not set");
}

/* ----------------------------------------------------------------------
   register a callback to a fix so it can manage atom-based arrays
   happens when fix is created
   flag = 0 for grow, 1 for restart
------------------------------------------------------------------------- */

void Atom::add_callback(int flag)
{
  int ifix;

  // find the fix
  // if find NULL ptr:
  //   it's this one, since it is deleted at this point in creation
  // if don't find NULL ptr:
  //   i is set to nfix = new one currently being added at end of list

  for (ifix = 0; ifix < modify->nfix; ifix++)
    if (modify->fix[ifix] == NULL) break;

  // add callback to lists, reallocating if necessary

  if (flag == 0) {
    if (nextra_grow == nextra_grow_max) {
      nextra_grow_max += DELTA_CALLBACK;
      extra_grow = (int *) 
	memory->srealloc(extra_grow,nextra_grow_max*sizeof(int),
			 "atom:extra_grow");
    }
    extra_grow[nextra_grow] = ifix;
    nextra_grow++;
  } else if (flag == 1) {
    if (nextra_restart == nextra_restart_max) {
      nextra_restart_max += DELTA_CALLBACK;
      extra_restart = (int *) 
	memory->srealloc(extra_restart,nextra_restart_max*sizeof(int),
			 "atom:extra_restart");
    }
    extra_restart[nextra_restart] = ifix;
    nextra_restart++;
  }
}

/* ----------------------------------------------------------------------
   unregister a callback to a fix
   happens when fix is deleted
   flag = 0 for grow, 1 for restart
------------------------------------------------------------------------- */

void Atom::delete_callback(char *id, int flag)
{
  int ifix;

  for (ifix = 0; ifix < modify->nfix; ifix++)
    if (strcmp(id,modify->fix[ifix]->id) == 0) break;

  // compact the list of callbacks

  if (flag == 0) {
    for (int i = ifix; i < nextra_grow-1; i++)
      extra_grow[i] = extra_grow[i+1];
    nextra_grow--;
  } else if (flag == 1) {
    for (int i = ifix; i < nextra_restart-1; i++)
      extra_restart[i] = extra_restart[i+1];
    nextra_restart--;
  }
}

/* ----------------------------------------------------------------------
   decrement ptrs in callback lists to fixes beyond the deleted ifix
   happens after fix is deleted
------------------------------------------------------------------------- */

void Atom::update_callback(int ifix)
{
  for (int i = 0; i < nextra_grow; i++)
    if (extra_grow[i] > ifix) extra_grow[i]--;
  for (int i = 0; i < nextra_restart; i++)
    if (extra_restart[i] > ifix) extra_restart[i]--;
}

/* ----------------------------------------------------------------------
   pack up xshake values into buf for communication
------------------------------------------------------------------------- */

void Atom::pack_shake(int n, double **xshake,
		      int *list, double *buf, int *pbc_flags)
{
  int i,j,m;

  m = 0;
  if (pbc_flags[0] == 0) {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = xshake[j][0];
      buf[m++] = xshake[j][1];
      buf[m++] = xshake[j][2];
    }
  } else {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = xshake[j][0] + pbc_flags[1]*domain->xprd;
      buf[m++] = xshake[j][1] + pbc_flags[2]*domain->yprd;
      buf[m++] = xshake[j][2] + pbc_flags[3]*domain->zprd;
    }
  }
}

/* ----------------------------------------------------------------------
   unpack xshake values from buf for communication
------------------------------------------------------------------------- */

void Atom::unpack_shake(int n, int first, double **xshake, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    xshake[i][0] = buf[m++];
    xshake[i][1] = buf[m++];
    xshake[i][2] = buf[m++];
  }
}

/* ----------------------------------------------------------------------
   pack up embedding values into buf for communication
------------------------------------------------------------------------- */

void Atom::pack_embed(int n, int *list, double *buf)
{
  int i,j,m;

  m = 0;
  for (i = 0; i < n; i++) {
    j = list[i];
    buf[m++] = fp[j];
  }
}

/* ----------------------------------------------------------------------
   unpack embedding values from buf for communication
------------------------------------------------------------------------- */

void Atom::unpack_embed(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) fp[i] = buf[m++];
}

/* ----------------------------------------------------------------------
   pack up density values into buf for communication
------------------------------------------------------------------------- */

void Atom::pack_reverse_rho(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) buf[m++] = rho[i];
}

/* ----------------------------------------------------------------------
   unpack density values from buf for communication
------------------------------------------------------------------------- */

void Atom::unpack_reverse_rho(int n, int *list, double *buf)
{
  int i,j,m;

  m = 0;
  for (i = 0; i < n; i++) {
    j = list[i];
    rho[j] += buf[m++];
  }
}
