/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#ifndef ATOM_H
#define ATOM_H

#include "system.h"

class Atom : public System {
 public:
  char *style;

  // atom counts

  int natoms;                   // total # of atoms in system, could be 0
  int nlocal,nghost;            // # on this proc, owned and ghost
  int nmax;                     // max # of owned+ghost in arrays

  // # of atom and topology types

  int ntypes,nbondtypes,nangletypes,ndihedraltypes,nimpropertypes;

  // molecular counts

  int nbonds,nangles,ndihedrals,nimpropers;
  int bond_per_atom,angle_per_atom,dihedral_per_atom,improper_per_atom;

  // options - set by each atom style

  int molecular;                      // 0 = atomic, 1 = molecular system
  int bonds_allow,angles_allow,dihedrals_allow,impropers_allow;    // 0/1
  int mass_require;                   // 0 = no mass types, 1 = yes
  int dipole_require;                 // 0 = no dipole moments, 1 = yes

  // communication sizes - set by each atom style

  int size_comm,size_reverse,size_border;
  int size_atom_valid,size_atom_actual;

  // arrays of length ntypes and flags if set

  double *mass;
  int *mass_flag;
  double *dipole;
  int *dipole_flag;

  // arrays common to all atom classes

  int *tag,*type,*mask,*image;
  double **x,**v,**f;

  // arrays specific to some atom classes

  double *q;
  double **mu;
  double **omega,**torque;
  double *rho,*fp;
  double *radius,*density,*rmass;
  double **phix,**phiv,**phia;

  int *molecule;

  int *num_bond;
  int **bond_type;
  int **bond_atom;

  int *num_angle;
  int **angle_type;
  int **angle_atom1,**angle_atom2,**angle_atom3;

  int *num_dihedral;
  int **dihedral_type;
  int **dihedral_atom1,**dihedral_atom2,**dihedral_atom3,**dihedral_atom4;

  int *num_improper;
  int **improper_type;
  int **improper_atom1,**improper_atom2,**improper_atom3,**improper_atom4;
 
  int maxspecial;
  int **nspecial,**special;

  // array for extra peratom info in restart file destined for fix & diag 

  double **extra;

  // ptr from global to local atoms

  int *map;

  // callback ptrs for atom arrays managed by fix classes

  int nextra_grow,nextra_restart;             // # of callbacks of each type
  int *extra_grow,*extra_restart;             // index of fix to callback to
  int nextra_grow_max,nextra_restart_max;     // size of callback lists
  int nextra_store;

  // functions common to all atom styles

  Atom(char *);
  virtual ~Atom();
  void init();

  void map_init();             // functions for global->local map array
  void map_clear();
  void map_set();
  void map_delete();

  void tag_extend();

  int parse_data(char *);
  int count_words(char *);

  virtual void unpack_vels(int, char *);     // can be overwritten by child
  void unpack_bonds(int, char *);
  void unpack_angles(int, char *);
  void unpack_dihedrals(int, char *);
  void unpack_impropers(int, char *);

  void allocate_type_arrays();
  void set_mass(char *);
  void set_mass(int, double);
  void set_mass(int, char **);
  void set_mass(double *);
  void check_mass();
  void set_dipole(char *);
  void set_dipole(int, char **);
  void set_dipole(double *);
  void check_dipole();

  void pack_shake(int, double **, int *, double *, int *);
  void unpack_shake(int, int, double **, double *);
  void pack_embed(int, int *, double *);
  void unpack_embed(int, int, double *);
  void pack_reverse_rho(int, int, double *);
  void unpack_reverse_rho(int, int *, double *);

  void add_callback(int);
  void delete_callback(char *, int);
  void update_callback(int);

  // optional child class functions

  virtual void init_style() {}

  // pure virtual functions, must be defined in child class

  virtual void grow(int) = 0;
  virtual int memory_usage() = 0;

  virtual void copy(int, int) = 0;
  virtual void pack_comm(int, int *, double *, int *) = 0;
  virtual void unpack_comm(int, int, double *) = 0;
  virtual void pack_reverse(int, int, double *) = 0;
  virtual void unpack_reverse(int, int *, double *) = 0;

  virtual int pack_border(int, double *, int *) = 0;
  virtual int unpack_border(int, double *) = 0;
  virtual int pack_exchange(int, double *) = 0;
  virtual int unpack_exchange(double *) = 0;
  virtual int pack_restart(int, double *) = 0;
  virtual int unpack_restart(double *) = 0;
  virtual int size_restart() = 0;

  virtual void unpack_data(int, char *) = 0;
  virtual void create_one(int, double, double, double) = 0;
};

#endif
