/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "mpi.h"
#include "atom_granular.h"
#include "domain.h"
#include "update.h"
#include "force.h"
#include "modify.h"
#include "fix.h"
#include "output.h"
#include "thermo.h"
#include "error.h"
#include "memory.h"
#include "comm.h"

#define DELTA 10000

/* ---------------------------------------------------------------------- */

AtomGranular::AtomGranular(char *style_in) : Atom(style_in)
{
  molecular = 0;
  bonds_allow = angles_allow = dihedrals_allow = impropers_allow = 0;
  mass_require = 0;
  dipole_require = 0;

  size_comm = 9;
  size_reverse = 6;
  size_border = 14;
  size_atom_valid = 7;

  tag = type = mask = image = NULL;
  x = v = f = NULL;
  phix = phiv = phia = NULL;
  radius = density = rmass = NULL;

  PI = 4.0*atan(1.0);
}

/* ---------------------------------------------------------------------- */

AtomGranular::~AtomGranular()
{
  memory->sfree(tag);
  memory->sfree(type);
  memory->sfree(mask);
  memory->sfree(image);
  memory->destroy_2d_double_array(x);
  memory->destroy_2d_double_array(v);
  memory->destroy_2d_double_array(f);
  memory->destroy_2d_double_array(phix);
  memory->destroy_2d_double_array(phiv);
  memory->destroy_2d_double_array(phia);
  memory->sfree(radius);
  memory->sfree(density);
  memory->sfree(rmass);
}

/* ---------------------------------------------------------------------- */

void AtomGranular::init_style()
{
  // insure thermo_gran is used
  // insure LJ units
  // insure 3d

  if (strcmp(output->thermo->style,"gran") != 0)
    error->all("Must use atom style granular with thermo style gran");
  if (strcmp(update->unit_style,"lj") != 0)
    error->all("Must use atom style granular with lj units");
  if (force->dimension == 2)
    error->all("Atom style granular must perform 3d simulations");
}

/* ----------------------------------------------------------------------
   grow atom arrays
   n = 0 grows arrays by DELTA
   n > 0 allocates arrays to size n
------------------------------------------------------------------------- */

void AtomGranular::grow(int n)
{
  if (n == 0) nmax += DELTA;
  else nmax = n;

  tag = (int *) memory->srealloc(tag,nmax*sizeof(int),"atom:tag");
  type = (int *) memory->srealloc(type,nmax*sizeof(int),"atom:type");
  mask = (int *) memory->srealloc(mask,nmax*sizeof(int),"atom:mask");
  image = (int *) memory->srealloc(image,nmax*sizeof(int),"atom:image");
  x = memory->grow_2d_double_array(x,nmax,3,"atom:x");
  v = memory->grow_2d_double_array(v,nmax,3,"atom:v");
  f = memory->grow_2d_double_array(f,nmax,3,"atom:f");
  phix = memory->grow_2d_double_array(phix,nmax,3,"atom:phix");
  phiv = memory->grow_2d_double_array(phiv,nmax,3,"atom:phiv");
  phia = memory->grow_2d_double_array(phia,nmax,3,"atom:phia");
  radius = (double *) memory->srealloc(radius,nmax*sizeof(double),
				       "atom:radius");
  density = (double *) memory->srealloc(density,nmax*sizeof(double),
					"atom:density");
  rmass = (double *) memory->srealloc(rmass,nmax*sizeof(double),
				      "atom:rmass");

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      modify->fix[extra_grow[iextra]]->grow_arrays(nmax);
}

/* ----------------------------------------------------------------------
   return # of bytes of allocated memory 
------------------------------------------------------------------------- */

int AtomGranular::memory_usage()
{
  int bytes = 0;

  bytes += 4 * nmax * sizeof(int);
  bytes += 6 * nmax*3 * sizeof(double);
  bytes += 3 * nmax * sizeof(double);

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      bytes += modify->fix[extra_grow[iextra]]->memory_usage(nmax);

  return bytes;
}

/* ---------------------------------------------------------------------- */

void AtomGranular::copy(int i, int j)
{
  tag[j] = tag[i];
  type[j] = type[i];
  mask[j] = mask[i];
  image[j] = image[i];
  x[j][0] = x[i][0];
  x[j][1] = x[i][1];
  x[j][2] = x[i][2];
  v[j][0] = v[i][0];
  v[j][1] = v[i][1];
  v[j][2] = v[i][2];
  phix[j][0] = phix[i][0];
  phix[j][1] = phix[i][1];
  phix[j][2] = phix[i][2];
  phiv[j][0] = phiv[i][0];
  phiv[j][1] = phiv[i][1];
  phiv[j][2] = phiv[i][2];
  radius[j] = radius[i];
  density[j] = density[i];
  rmass[j] = rmass[i];

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      modify->fix[extra_grow[iextra]]->copy_arrays(i,j);
}

/* ---------------------------------------------------------------------- */

void AtomGranular::pack_comm(int n, int *list, double *buf, int *pbc_flags)
{
  int i,j,m;

  m = 0;
  if (pbc_flags[0] == 0) {
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0];
      buf[m++] = x[j][1];
      buf[m++] = x[j][2];
      buf[m++] = v[j][0];
      buf[m++] = v[j][1];
      buf[m++] = v[j][2];
      buf[m++] = phiv[j][0];
      buf[m++] = phiv[j][1];
      buf[m++] = phiv[j][2];
    }
  } else {
    double xprd = domain->xprd;
    double yprd = domain->yprd;
    double zprd = domain->zprd;
    for (i = 0; i < n; i++) {
      j = list[i];
      buf[m++] = x[j][0] + pbc_flags[1]*xprd;
      buf[m++] = x[j][1] + pbc_flags[2]*yprd;
      buf[m++] = x[j][2] + pbc_flags[3]*zprd;
      buf[m++] = v[j][0];
      buf[m++] = v[j][1];
      buf[m++] = v[j][2];
      buf[m++] = phiv[j][0];
      buf[m++] = phiv[j][1];
      buf[m++] = phiv[j][2];
    }
  }
}

/* ---------------------------------------------------------------------- */

void AtomGranular::unpack_comm(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    x[i][0] = buf[m++];
    x[i][1] = buf[m++];
    x[i][2] = buf[m++];
    v[i][0] = buf[m++];
    v[i][1] = buf[m++];
    v[i][2] = buf[m++];
    phiv[i][0] = buf[m++];
    phiv[i][1] = buf[m++];
    phiv[i][2] = buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

void AtomGranular::pack_reverse(int n, int first, double *buf)
{
  int i,m,last;

  m = 0;
  last = first + n;
  for (i = first; i < last; i++) {
    buf[m++] = f[i][0];
    buf[m++] = f[i][1];
    buf[m++] = f[i][2];
    buf[m++] = phia[i][0];
    buf[m++] = phia[i][1];
    buf[m++] = phia[i][2];
  }
}

/* ---------------------------------------------------------------------- */

void AtomGranular::unpack_reverse(int n, int *list, double *buf)
{
  int i,j,m;

  m = 0;
  for (i = 0; i < n; i++) {
    j = list[i];
    f[j][0] += buf[m++];
    f[j][1] += buf[m++];
    f[j][2] += buf[m++];
    phia[j][0] += buf[m++];
    phia[j][1] += buf[m++];
    phia[j][2] += buf[m++];
  }
}

/* ---------------------------------------------------------------------- */

int AtomGranular::pack_border(int i, double *buf, int *pbc_flags)
{
  int m = 0;
  if (pbc_flags[0] == 0) {
    buf[m++] = x[i][0];
    buf[m++] = x[i][1];
    buf[m++] = x[i][2];
  } else {
    buf[m++] = x[i][0] + pbc_flags[1]*domain->xprd;
    buf[m++] = x[i][1] + pbc_flags[2]*domain->yprd;
    buf[m++] = x[i][2] + pbc_flags[3]*domain->zprd;
  }
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];
  buf[m++] = phiv[i][0];
  buf[m++] = phiv[i][1];
  buf[m++] = phiv[i][2];
  buf[m++] = radius[i];
  buf[m++] = density[i];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomGranular::unpack_border(int i, double *buf)
{
  if (i == nmax) grow(0);

  int m = 0;
  x[i][0] = buf[m++];
  x[i][1] = buf[m++];
  x[i][2] = buf[m++];
  v[i][0] = buf[m++];
  v[i][1] = buf[m++];
  v[i][2] = buf[m++];
  phiv[i][0] = buf[m++];
  phiv[i][1] = buf[m++];
  phiv[i][2] = buf[m++];
  radius[i] = buf[m++];
  density[i] = buf[m++];
  rmass[i] = 4.0*PI/3.0 * radius[i]*radius[i]*radius[i] * density[i];
  tag[i] = static_cast<int> (buf[m++]);
  type[i] = static_cast<int> (buf[m++]);
  mask[i] = static_cast<int> (buf[m++]);

  return m;
}

/* ----------------------------------------------------------------------
   pack all atom quantities for shipping to another proc
   xyz must be 1st 3 values, so that comm::exchange can test on them
------------------------------------------------------------------------- */

int AtomGranular::pack_exchange(int i, double *buf)
{
  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];
  buf[m++] = phix[i][0];
  buf[m++] = phix[i][1];
  buf[m++] = phix[i][2];
  buf[m++] = phiv[i][0];
  buf[m++] = phiv[i][1];
  buf[m++] = phiv[i][2];
  buf[m++] = radius[i];
  buf[m++] = density[i];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (image[i]);

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      m += modify->fix[extra_grow[iextra]]->pack_exchange(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomGranular::unpack_exchange(double *buf)
{
  if (nlocal == nmax) grow(0);

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];
  phix[nlocal][0] = buf[m++];
  phix[nlocal][1] = buf[m++];
  phix[nlocal][2] = buf[m++];
  phiv[nlocal][0] = buf[m++];
  phiv[nlocal][1] = buf[m++];
  phiv[nlocal][2] = buf[m++];
  radius[nlocal] = buf[m++];
  density[nlocal] = buf[m++];
  rmass[nlocal] = 4.0*PI/3.0 * 
    radius[nlocal]*radius[nlocal]*radius[nlocal] * density[nlocal];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);

  if (nextra_grow)
    for (int iextra = 0; iextra < nextra_grow; iextra++) 
      m += modify->fix[extra_grow[iextra]]->unpack_exchange(nlocal,&buf[m]);

  nlocal++;
  return m;
}

/* ---------------------------------------------------------------------- */

int AtomGranular::size_restart()
{
  int i;

  int n = 19 * nlocal;

  if (nextra_restart)
    for (int iextra = 0; iextra < nextra_restart; iextra++) 
      for (i = 0; i < nlocal; i++)
	n += modify->fix[extra_restart[iextra]]->size_restart(i);

  return n;
}

/* ----------------------------------------------------------------------
   pack all atom quantities for restart file including extra quantities
   xyz must be 1st 3 values, so that read_restart can test on them
------------------------------------------------------------------------- */

int AtomGranular::pack_restart(int i, double *buf)
{
  int m = 1;
  buf[m++] = x[i][0];
  buf[m++] = x[i][1];
  buf[m++] = x[i][2];
  buf[m++] = v[i][0];
  buf[m++] = v[i][1];
  buf[m++] = v[i][2];
  buf[m++] = phix[i][0];
  buf[m++] = phix[i][1];
  buf[m++] = phix[i][2];
  buf[m++] = phiv[i][0];
  buf[m++] = phiv[i][1];
  buf[m++] = phiv[i][2];
  buf[m++] = radius[i];
  buf[m++] = density[i];
  buf[m++] = static_cast<double> (tag[i]);
  buf[m++] = static_cast<double> (type[i]);
  buf[m++] = static_cast<double> (mask[i]);
  buf[m++] = static_cast<double> (image[i]);

  if (nextra_restart)
    for (int iextra = 0; iextra < nextra_restart; iextra++) 
      m += modify->fix[extra_restart[iextra]]->pack_restart(i,&buf[m]);

  buf[0] = m;
  return m;
}

/* ----------------------------------------------------------------------
   unpack all atom quantities from restart file including extra quantities
------------------------------------------------------------------------- */

int AtomGranular::unpack_restart(double *buf)
{
  if (nlocal == nmax) {
    grow(0);
    if (nextra_store)
      extra = memory->grow_2d_double_array(extra,nmax,nextra_store,
					   "atom:extra");
  }

  int m = 1;
  x[nlocal][0] = buf[m++];
  x[nlocal][1] = buf[m++];
  x[nlocal][2] = buf[m++];
  v[nlocal][0] = buf[m++];
  v[nlocal][1] = buf[m++];
  v[nlocal][2] = buf[m++];
  phix[nlocal][0] = buf[m++];
  phix[nlocal][1] = buf[m++];
  phix[nlocal][2] = buf[m++];
  phiv[nlocal][0] = buf[m++];
  phiv[nlocal][1] = buf[m++];
  phiv[nlocal][2] = buf[m++];
  radius[nlocal] = buf[m++];
  density[nlocal] = buf[m++];
  rmass[nlocal] = 4.0*PI/3.0 * 
    radius[nlocal]*radius[nlocal]*radius[nlocal] * density[nlocal];
  tag[nlocal] = static_cast<int> (buf[m++]);
  type[nlocal] = static_cast<int> (buf[m++]);
  mask[nlocal] = static_cast<int> (buf[m++]);
  image[nlocal] = static_cast<int> (buf[m++]);

  if (nextra_store) {
    int size = static_cast<int> (buf[0]) - m;
    for (int i = 0; i < size; i++) extra[nlocal][i] = buf[m++];
  }
  
  nlocal++;
  return m;
}

/* ----------------------------------------------------------------------
   unpack n lines of atoms from buf
------------------------------------------------------------------------- */

void AtomGranular::unpack_data(int n, char *buf)
{
  int tagtmp,typetmp,imagetmp;
  double radtmp,denstmp,xtmp,ytmp,ztmp;
  int imagex,imagey,imagez;
  char *next;

  double subxlo = domain->subxlo;
  double subxhi = domain->subxhi;
  double subylo = domain->subylo;
  double subyhi = domain->subyhi;
  double subzlo = domain->subzlo;
  double subzhi = domain->subzhi;

  for (int i = 0; i < n; i++) {

    next = strchr(buf,'\n');
    *next = '\0';

    if (size_atom_actual == 7) {
      sscanf(buf,"%d %d %lg %lg %lg %lg %lg",
	     &tagtmp,&typetmp,&radtmp,&denstmp,&xtmp,&ytmp,&ztmp);
      imagetmp = 500500500;
    } else {
      sscanf(buf,"%d %d %lg %lg %lg %lg %lg %d %d %d",
	     &tagtmp,&typetmp,&radtmp,&denstmp,
	     &xtmp,&ytmp,&ztmp,&imagex,&imagey,&imagez);
      imagetmp = 1000000 * (500+imagez) + 1000 * (500+imagey) + 500+imagex;
    }

    domain->remap(xtmp,ytmp,ztmp,imagetmp);
    if (xtmp >= subxlo && xtmp < subxhi &&
	ytmp >= subylo && ytmp < subyhi &&
	ztmp >= subzlo && ztmp < subzhi) {
      if (nlocal == nmax) grow(0);
      tag[nlocal] = tagtmp;
      type[nlocal] = typetmp;
      radius[nlocal] = 0.5*radtmp;
      density[nlocal] = denstmp;
      rmass[nlocal] = 4.0*PI/3.0 * 
	radius[nlocal]*radius[nlocal]*radius[nlocal] * density[nlocal];
      x[nlocal][0] = xtmp;
      x[nlocal][1] = ytmp;
      x[nlocal][2] = ztmp;
      mask[nlocal] = 1;
      image[nlocal] = imagetmp;
      v[nlocal][0] = 0.0;
      v[nlocal][1] = 0.0;
      v[nlocal][2] = 0.0;
      phix[nlocal][0] = 0.0;
      phix[nlocal][1] = 0.0;
      phix[nlocal][2] = 0.0;
      phiv[nlocal][0] = 0.0;
      phiv[nlocal][1] = 0.0;
      phiv[nlocal][2] = 0.0;
      nlocal++;
    }

    buf = next + 1;
  }
}

/* ----------------------------------------------------------------------
   create one atom of type itype at px,py,pz
------------------------------------------------------------------------- */

void AtomGranular::create_one(int itype, double px, double py, double pz)
{
  if (nlocal == nmax) grow(0);
  tag[nlocal] = 0;
  type[nlocal] = itype;
  radius[nlocal] = 0.5;
  density[nlocal] = 1.0;
  rmass[nlocal] = 4.0*PI/3.0 *
    radius[nlocal]*radius[nlocal]*radius[nlocal] * density[nlocal];
  x[nlocal][0] = px;
  x[nlocal][1] = py;
  x[nlocal][2] = pz;
  mask[nlocal] = 1;
  image[nlocal] = 500500500;
  v[nlocal][0] = 0.0;
  v[nlocal][1] = 0.0;
  v[nlocal][2] = 0.0;
  phix[nlocal][0] = 0.0;
  phix[nlocal][1] = 0.0;
  phix[nlocal][2] = 0.0;
  phiv[nlocal][0] = 0.0;
  phiv[nlocal][1] = 0.0;
  phiv[nlocal][2] = 0.0;
  nlocal++;
}

/* ----------------------------------------------------------------------
   unpack vels & phis specific to granular data file
------------------------------------------------------------------------- */

void AtomGranular::unpack_vels(int n, char *buf)
{
  int m,tagtmp;
  double vxtmp,vytmp,vztmp,phivxtmp,phivytmp,phivztmp;
  char *next;

  for (int i = 0; i < n; i++) {
    next = strchr(buf,'\n');
    *next = '\0';
    sscanf(buf,"%d %lg %lg %lg %lg %lg %lg",
	   &tagtmp,&vxtmp,&vytmp,&vztmp,
	   &phivxtmp,&phivytmp,&phivztmp);
    if ((m = map[tagtmp]) >= 0) {
      v[m][0] = vxtmp;
      v[m][1] = vytmp;
      v[m][2] = vztmp;
      phiv[m][0] = phivxtmp;
      phiv[m][1] = phivytmp;
      phiv[m][2] = phivztmp;
    }
    buf = next + 1;
  }
}
