/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

// LAMMPS main driver

#include "string.h"
#include "system.h"
#include "mpi.h"
#include "input.h"
#include "create_box.h"
#include "create_atoms.h"
#include "delete_atoms.h"
#include "displace_atoms.h"
#include "delete_bonds.h"
#include "set_type.h"
#include "read_data.h"
#include "read_restart.h"
#include "replicate.h"
#include "write_restart.h"
#include "velocity.h"
#include "update.h"
#include "domain.h"
#include "comm.h"
#include "integrate.h"
#include "finish.h"
#include "minimize.h"
#include "temper.h"

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
  System system;
  system.open(&argc,&argv);
  system.create();

  Input input;
  input.initvar(argc,argv);
  char *command;

  while (command = input.next()) {

    if (strcmp(command,"create_box") == 0) {
      CreateBox create_box;
      create_box.create(input.narg,input.arg);

    } else if (strcmp(command,"create_atoms") == 0) {
      CreateAtoms create_atoms;
      create_atoms.create(input.narg,input.arg);

    } else if (strcmp(command,"delete_atoms") == 0) {
      DeleteAtoms delete_atoms;
      delete_atoms.destroy(input.narg,input.arg);

    } else if (strcmp(command,"displace_atoms") == 0) {
      DisplaceAtoms displace_atoms;
      displace_atoms.move(input.narg,input.arg);

    } else if (strcmp(command,"delete_bonds") == 0) {
      DeleteBonds delete_bonds;
      delete_bonds.destroy(input.narg,input.arg);

    } else if (strcmp(command,"set_type") == 0) {
      SetType set_type;
      set_type.set(input.narg,input.arg);

    } else if (strcmp(command,"read_data") == 0) {
      ReadData read_data;
      read_data.read(input.arg[0]);

    } else if (strcmp(command,"read_restart") == 0) {
      ReadRestart read_restart;
      read_restart.read(input.arg[0]);

    } else if (strcmp(command,"replicate") == 0) {
      Replicate replicate;
      replicate.replicate(input.arg[0],input.arg[1],input.arg[2]);

    } else if (strcmp(command,"velocity") == 0) {
      Velocity velocity;
      velocity.generate(input.narg,input.arg);

    } else if (strcmp(command,"write_restart") == 0) {
      system.init();
      system.domain->pbc();          // enforce atoms in simulation box
      system.comm->setup();
      system.comm->exchange();
      WriteRestart write_restart;
      write_restart.write(input.arg[0]);

    } else if (strcmp(command,"run") == 0) {
      system.init();
      system.update->integrate->run();
      Finish finish;
      finish.end();

    } else if (strcmp(command,"minimize") == 0) {
      system.init();
      system.update->minimize->run();
      Finish finish;
      finish.end();

    } else if (strcmp(command,"temper") == 0) {
      system.init();
      Temper temper(input.narg,input.arg);
      temper.run();
      Finish finish;
      finish.end();

    } else if (strcmp(command,"clear") == 0) {
      system.destroy();
      system.create();
    }

  }

  system.destroy();
  system.close();
}
