/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "stdio.h"
#include "string.h"
#include "mpi.h"
#include "pair.h"
#include "atom.h"
#include "force.h"
#include "update.h"
#include "integrate.h"
#include "error.h"
#include <climits>

#define GEOMETRIC  0
#define ARITHMETIC 1
#define SIXTHPOWER 2

/* ---------------------------------------------------------------------- */

Pair::Pair()
{
  eng_vdwl = eng_coul = 0.0;
  allocated = 0;
  offset_flag = 0;
  mix = GEOMETRIC;
  ntablebits = 12;
}

/* ----------------------------------------------------------------------
   modify parameters of the pair style 
------------------------------------------------------------------------- */

void Pair::modify_params(int narg, char **arg)
{
  if (narg == 0) error->all("Illegal pair_modify command");

  int iarg = 0;
  while (iarg < narg) {
    if (strcmp(arg[iarg],"mix") == 0) {
      if (iarg+2 > narg) error->all("Illegal pair_modify command");
      if (strcmp(arg[iarg+1],"geometric") == 0) mix = GEOMETRIC;
      else if (strcmp(arg[iarg+1],"arithmetic") == 0) mix = ARITHMETIC;
      else if (strcmp(arg[iarg+1],"sixthpower") == 0) mix = SIXTHPOWER;
      else error->all("Incorrect args in pair_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"shift") == 0) {
      if (iarg+2 > narg) error->all("Illegal pair_modify command");
      if (strcmp(arg[iarg+1],"yes") == 0) offset_flag = 1;
      else if (strcmp(arg[iarg+1],"no") == 0) offset_flag = 0;
      else error->all("Incorrect args in pair_modify command");
      iarg += 2;
    } else if (strcmp(arg[iarg],"table") == 0) {
      if (iarg+2 > narg) error->all("Illegal pair_modify command");
      ntablebits = atoi(arg[iarg+1]);
      if (ntablebits > sizeof(float)*CHAR_BIT) 
        error->all("Too many total bits for bitmapped lookup table");
      iarg += 2;
    } else error->all("Incorrect args in pair_modify command");
  }
}

/* ----------------------------------------------------------------------
   compute pair virial via pair own/ghost forces 
------------------------------------------------------------------------- */

void Pair::virial_compute()
{
  double **x = atom->x;
  double **f_pair = update->integrate->f_pair;
  int nall = atom->nlocal + atom->nghost;

  // sum over own & ghost atoms

  for (int i = 0; i < nall; i++) {
    virial[0] += f_pair[i][0]*x[i][0];
    virial[1] += f_pair[i][1]*x[i][1];
    virial[2] += f_pair[i][2]*x[i][2];
    virial[3] += f_pair[i][1]*x[i][0];
    virial[4] += f_pair[i][2]*x[i][0];
    virial[5] += f_pair[i][2]*x[i][1];
  }

  // add pair forces into total force

  double **f = atom->f;

  for (int i = 0; i < nall; i++) {
    f[i][0] += f_pair[i][0];
    f[i][1] += f_pair[i][1];
    f[i][2] += f_pair[i][2];
  }
}

/* ----------------------------------------------------------------------
   write a table of pair potential values to a file
------------------------------------------------------------------------- */

/*

void Pair::write_file(int narg, char **arg)
{
  if (narg < 6) error->all("Incorrect args in pair_write command");

  // initialize before evaluating to insure all coeffs are set

  init();

  // parse arguments

  int n = atoi(arg[0]);
  double inner = atof(arg[1]);
  double outer = atof(arg[2]);

  int me;
  MPI_Comm_rank(world,&me);
  FILE *fp;
  if (me == 0) {
    fp = fopen(arg[3],"w");
    if (fp == NULL) error->one("Cannot open pair_write file");
  }

  int itype = atoi(arg[4]);
  int jtype = atoi(arg[5]);
  if (itype > jtype) error->all("Incorrect args in pair_write command");
  if (outer*outer > cutsq[itype][jtype]) 
    error->all("Pair_write cutoff is too large");

  int qflag = 0;
  if (strcmp(force->pair_style,"lj/cut/coul/cut") == 0 ||
      strcmp(force->pair_style,"lj/cut/coul/debye") == 0 ||
      strcmp(force->pair_style,"lj/cut/coul/long") == 0 ||
      strcmp(force->pair_style,"lj/charmm/coul/charmm") == 0 ||
      strcmp(force->pair_style,"lj/charmm/coul/charmm/implicit") == 0 ||
      strcmp(force->pair_style,"lj/charmm/coul/long") == 0)
    qflag = 1;

  double qi,qj;
  if (qflag) {
    if (narg != 8) error->all("Incorrect args in pair_write command");
    qi = atof(arg[6]);
    qj = atof(arg[7]);
  }

  double r,e,f;
  for (int i = 0; i < n; i++) {
    r = inner + (outer-inner) * i/(n-1);
    if (qflag) {
      e = esingle(r,itype,jtype,qi,qj);
      f = fsingle(r,itype,jtype,qi,qj);
    } else {
      e = esingle(r,itype,jtype);
      f = fsingle(r,itype,jtype);
    }
    if (me == 0) fprintf(fp,"%d %g %g %g\n",i+1,r,e,f);
  }

  if (me == 0) fclose(fp);
}

*/
