/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "stdio.h"
#include "string.h"
#include "pair_lj_expand.h"
#include "atom.h"
#include "comm.h"
#include "force.h"
#include "update.h"
#include "integrate.h"
#include "memory.h"
#include "neighbor.h"
#include "error.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define GEOMETRIC  0
#define ARITHMETIC 1
#define SIXTHPOWER 2

/* ---------------------------------------------------------------------- */

PairLJExpand::~PairLJExpand()
{
  if (allocated) {
    memory->destroy_2d_int_array(setflag);
    memory->destroy_2d_double_array(cutsq);

    memory->destroy_2d_double_array(cut_lj);
    memory->destroy_2d_double_array(cut_ljsq);
    memory->destroy_2d_double_array(epsilon);
    memory->destroy_2d_double_array(sigma);
    memory->destroy_2d_double_array(shift);
    memory->destroy_2d_double_array(lj1);
    memory->destroy_2d_double_array(lj2);
    memory->destroy_2d_double_array(lj3);
    memory->destroy_2d_double_array(lj4);
    memory->destroy_2d_double_array(offset);
  }
}

/* ---------------------------------------------------------------------- */

void PairLJExpand::compute(int eflag, int vflag)
{
  int i,j,k,numneigh,itype,jtype;
  double xtmp,ytmp,ztmp,delx,dely,delz;
  double rsq,r2inv,r6inv,forcelj,fforce,factor,phi;
  double r,rshift,rshiftsq;
  int *neighs;
  double **f;

  eng_vdwl = 0.0;
  if (vflag) for (i = 0; i < 6; i++) virial[i] = 0.0;

  if (vflag == 2) f = update->integrate->f_pair;
  else f = atom->f;
  double **x = atom->x;
  int *type = atom->type;
  int nlocal = atom->nlocal;
  int nall = atom->nlocal + atom->nghost;
  double *special_lj = force->special_lj;
  int newton_pair = force->newton_pair;

  // loop over neighbors of my atoms

  for (i = 0; i < nlocal; i++) {
    xtmp = x[i][0];
    ytmp = x[i][1];
    ztmp = x[i][2];
    itype = type[i];
    neighs = neighbor->firstneigh[i];
    numneigh = neighbor->numneigh[i];

    for (k = 0; k < numneigh; k++) {
      j = neighs[k];

      if (j < nall) factor = 1.0;
      else {
	factor = special_lj[j/nall];
	j = j % nall;
      }

      delx = xtmp - x[j][0];
      dely = ytmp - x[j][1];
      delz = ztmp - x[j][2];
      rsq = delx*delx + dely*dely + delz*delz;
      jtype = type[j];

      if (rsq < cut_ljsq[itype][jtype]) {
	r = sqrt(rsq);
	rshift = r - shift[itype][jtype];
	rshiftsq = rshift*rshift;
	r2inv = 1.0/rshiftsq;
	r6inv = r2inv*r2inv*r2inv;
	forcelj = r6inv * (lj1[itype][jtype]*r6inv - lj2[itype][jtype]);
	fforce = factor*forcelj/rshift/r;

	f[i][0] += delx*fforce;
	f[i][1] += dely*fforce;
	f[i][2] += delz*fforce;
	if (newton_pair || j < nlocal) {
	  f[j][0] -= delx*fforce;
	  f[j][1] -= dely*fforce;
	  f[j][2] -= delz*fforce;
	}

	if (eflag) {
	  phi = r6inv*(lj3[itype][jtype]*r6inv-lj4[itype][jtype]) -
	    offset[itype][jtype];
	  if (newton_pair || j < nlocal) eng_vdwl += factor*phi;
	  else eng_vdwl += 0.5*factor*phi;
	}

	if (vflag == 1) {
	  if (newton_pair || j < nlocal) {
	    virial[0] += delx*delx*fforce;
	    virial[1] += dely*dely*fforce;
	    virial[2] += delz*delz*fforce;
	    virial[3] += delx*dely*fforce;
	    virial[4] += delx*delz*fforce;
	    virial[5] += dely*delz*fforce;
	  } else {
	    virial[0] += 0.5*delx*delx*fforce;
	    virial[1] += 0.5*dely*dely*fforce;
	    virial[2] += 0.5*delz*delz*fforce;
	    virial[3] += 0.5*delx*dely*fforce;
	    virial[4] += 0.5*delx*delz*fforce;
	    virial[5] += 0.5*dely*delz*fforce;
	  }
	}
      }
    }
  }
  if (vflag == 2) virial_compute();
}

/* ----------------------------------------------------------------------
   allocate all arrays 
------------------------------------------------------------------------- */

void PairLJExpand::allocate()
{
  allocated = 1;
  int n = atom->ntypes;

  setflag = memory->create_2d_int_array(n+1,n+1,"pair:setflag");
  for (int i = 1; i <= n; i++)
    for (int j = i; j <= n; j++)
      setflag[i][j] = 0;

  cutsq = memory->create_2d_double_array(n+1,n+1,"pair:cutsq");

  cut_lj = memory->create_2d_double_array(n+1,n+1,"pair:cut_lj");
  cut_ljsq = memory->create_2d_double_array(n+1,n+1,"pair:cut_ljsq");
  epsilon = memory->create_2d_double_array(n+1,n+1,"pair:epsilon");
  sigma = memory->create_2d_double_array(n+1,n+1,"pair:sigma");
  shift = memory->create_2d_double_array(n+1,n+1,"pair:shift");
  lj1 = memory->create_2d_double_array(n+1,n+1,"pair:lj1");
  lj2 = memory->create_2d_double_array(n+1,n+1,"pair:lj2");
  lj3 = memory->create_2d_double_array(n+1,n+1,"pair:lj3");
  lj4 = memory->create_2d_double_array(n+1,n+1,"pair:lj4");
  offset = memory->create_2d_double_array(n+1,n+1,"pair:offset");
}

/* ----------------------------------------------------------------------
   global settings 
------------------------------------------------------------------------- */

void PairLJExpand::settings(int narg, char **arg)
{
  if (narg != 1) error->all("Incorrect args in pair_style command");

  cut_lj_global = atof(arg[0]);

  // reset cutoffs that were previously set from data file

  if (allocated) {
    int i,j;
    for (i = 1; i <= atom->ntypes; i++)
      for (j = i+1; j <= atom->ntypes; j++)
	if (setflag[i][j] == 1) cut_lj[i][j] = cut_lj_global;
  }
}

/* ----------------------------------------------------------------------
   set i,j coeffs from one line in input script 
------------------------------------------------------------------------- */

void PairLJExpand::set_coeff(int narg, char **arg)
{
  if (narg != 6) error->all("Incorrect args in pair_coeff command");
  if (!allocated) allocate();

  int ilo,ihi,jlo,jhi;
  force->bounds(arg[0],atom->ntypes,ilo,ihi);
  force->bounds(arg[1],atom->ntypes,jlo,jhi);

  double epsilon_one = atof(arg[2]);
  double sigma_one = atof(arg[3]);
  double shift_one = atof(arg[4]);
  double cut_lj_one = atof(arg[5]);

  int count = 0;
  for (int i = ilo; i <= ihi; i++) {
    for (int j = MAX(jlo,i); j <= jhi; j++) {
      epsilon[i][j] = epsilon_one;
      sigma[i][j] = sigma_one;
      shift[i][j] = shift_one;
      cut_lj[i][j] = cut_lj_one;
      setflag[i][j] = 2;
      count++;
    }
  }

  if (count == 0) error->all("Incorrect args in pair_coeff command");
}

/* ----------------------------------------------------------------------
   set i,i coeffs from one line in read_data file 
------------------------------------------------------------------------- */

void PairLJExpand::set_coeff(char *str)
{
  if (!allocated) allocate();

  int i;
  double epsilon_one,sigma_one,shift_one;
  sscanf(str,"%d %lg %lg %lg",&i,&epsilon_one,&sigma_one,&shift_one);

  epsilon[i][i] = epsilon_one;
  sigma[i][i] = sigma_one;
  shift[i][i] = shift_one;
  cut_lj[i][i] = cut_lj_global;
  setflag[i][i] = 1;
}

/* ----------------------------------------------------------------------
   error check and initialize all values needed for force computation 
------------------------------------------------------------------------- */

void PairLJExpand::init()
{
  int i,j;

  // all i,i terms must be set at this point

  if (!allocated) error->all("Pair coeffs are not set");
  for (i = 1; i <= atom->ntypes; i++)
    if (setflag[i][i] == 0) error->all("All pair coeffs are not set");

  // perform mixing of i < j terms
  // do not mix if term has been explicitly set in input script
  // mix will create epsilon, sigma, cut_lj (mixed same as sigma)

  for (i = 1; i <= atom->ntypes; i++)
    for (j = i+1; j <= atom->ntypes; j++)

      if (setflag[i][j] == 0) {
	if (mix == GEOMETRIC) {
	  epsilon[i][j] = sqrt(epsilon[i][i]*epsilon[j][j]);
	  sigma[i][j] = sqrt(sigma[i][i]*sigma[j][j]);
	  shift[i][j] = 0.5 * (shift[i][i] + shift[j][j]);
	  cut_lj[i][j] = sqrt(cut_lj[i][i]*cut_lj[j][j]);
	} else if (mix == ARITHMETIC) {
	  epsilon[i][j] = sqrt(epsilon[i][i]*epsilon[j][j]);
	  sigma[i][j] = 0.5 * (sigma[i][i] + sigma[j][j]);
	  shift[i][j] = 0.5 * (shift[i][i] + shift[j][j]);
	  cut_lj[i][j] = 0.5 * (cut_lj[i][i] + cut_lj[j][j]);
	} else if (mix == SIXTHPOWER) {
	  epsilon[i][j] = 2.0 * sqrt(epsilon[i][i]*epsilon[j][j]) *
	    pow(sigma[i][i],3.0) * pow(sigma[j][j],3.0) /
	    (pow(sigma[i][i],6.0) * pow(sigma[j][j],6.0));
	  sigma[i][j] = pow((0.5 * (pow(sigma[i][i],6.0) + 
				    pow(sigma[j][j],6.0))),1.0/6.0);
	  shift[i][j] = 0.5 * (shift[i][i] + shift[j][j]);
	  cut_lj[i][j] = pow((0.5 * (pow(cut_lj[i][i],6.0) + 
				    pow(cut_lj[j][j],6.0))),1.0/6.0);
	}
      }

  // now have epsilon, sigma, shift, cut_lj for all i <= j
  // cutsq = cutoff for each i,j pair
  // cutforce = longest cutoff for any i,j pair
  // cut_ljsq, lj1234, shift, offset = arrays used in force & energy

  double cut,ratio;
  cutforce = 0.0;
  for (i = 1; i <= atom->ntypes; i++)
    for (j = i; j <= atom->ntypes; j++) {
      cut = cut_lj[i][j] + shift[i][j];
      cutsq[i][j] = cut*cut;
      cutsq[j][i] = cutsq[i][j];
      cutforce = MAX(cutforce,cut);

      cut_ljsq[i][j] = (cut_lj[i][j]+shift[i][j]) * (cut_lj[i][j]+shift[i][j]);
      lj1[i][j] = 48.0 * epsilon[i][j] * pow(sigma[i][j],12.0);
      lj2[i][j] = 24.0 * epsilon[i][j] * pow(sigma[i][j],6.0);
      lj3[i][j] = 4.0 * epsilon[i][j] * pow(sigma[i][j],12.0);
      lj4[i][j] = 4.0 * epsilon[i][j] * pow(sigma[i][j],6.0);
     
      if (offset_flag) {
	ratio = sigma[i][j] / cut_lj[i][j];
	offset[i][j] = 4.0 * epsilon[i][j] * 
	  (pow(ratio,12.0) - pow(ratio,6.0));
      } else offset[i][j] = 0.0;

      cut_ljsq[j][i] = cut_ljsq[i][j];
      lj1[j][i] = lj1[i][j];
      lj2[j][i] = lj2[i][j];
      lj3[j][i] = lj3[i][j];
      lj4[j][i] = lj4[i][j];
      shift[j][i] = shift[i][j];
      offset[j][i] = offset[i][j];
    }
}

/* ----------------------------------------------------------------------
   proc 0 writes out coeffs to restart file 
------------------------------------------------------------------------- */

void PairLJExpand::write_restart(FILE *fp)
{
  int i,j;

  fwrite(&cut_lj_global,sizeof(double),1,fp);
  fwrite(&offset_flag,sizeof(int),1,fp);
  fwrite(&mix,sizeof(int),1,fp);

  for (i = 1; i <= atom->ntypes; i++)
    for (j = i; j <= atom->ntypes; j++) {
      fwrite(&setflag[i][j],sizeof(int),1,fp);
      if (setflag[i][j]) {
	fwrite(&epsilon[i][j],sizeof(double),1,fp);
	fwrite(&sigma[i][j],sizeof(double),1,fp);
	fwrite(&shift[i][j],sizeof(double),1,fp);
	fwrite(&cut_lj[i][j],sizeof(double),1,fp);
      }
    }
}

/* ----------------------------------------------------------------------
   proc 0 reads coeffs from restart file, bcasts them 
------------------------------------------------------------------------- */

void PairLJExpand::read_restart(FILE *fp)
{
  int i,j;

  allocate();

  int me = comm->me;

  if (me == 0) {
    fread(&cut_lj_global,sizeof(double),1,fp);
    fread(&offset_flag,sizeof(int),1,fp);
    fread(&mix,sizeof(int),1,fp);
  }
  MPI_Bcast(&cut_lj_global,1,MPI_DOUBLE,0,world);
  MPI_Bcast(&offset_flag,1,MPI_INT,0,world);
  MPI_Bcast(&mix,1,MPI_INT,0,world);

  for (i = 1; i <= atom->ntypes; i++)
    for (j = i; j <= atom->ntypes; j++) {
      if (me == 0) fread(&setflag[i][j],sizeof(int),1,fp);
      MPI_Bcast(&setflag[i][j],1,MPI_INT,0,world);
      if (setflag[i][j]) {
	if (me == 0) {
	  fread(&epsilon[i][j],sizeof(double),1,fp);
	  fread(&sigma[i][j],sizeof(double),1,fp);
	  fread(&shift[i][j],sizeof(double),1,fp);
	  fread(&cut_lj[i][j],sizeof(double),1,fp);
	}
	MPI_Bcast(&epsilon[i][j],1,MPI_DOUBLE,0,world);
	MPI_Bcast(&sigma[i][j],1,MPI_DOUBLE,0,world);
	MPI_Bcast(&shift[i][j],1,MPI_DOUBLE,0,world);
	MPI_Bcast(&cut_lj[i][j],1,MPI_DOUBLE,0,world);
      }
    }
}

/* ---------------------------------------------------------------------- */

/*
double PairLJExpand::esingle(double r, int itype, int jtype)
{
  double rshift = r - shift[itype][jtype];
  double rshiftsq = rshift*rshift;
  double r2inv = 1.0/rshiftsq;
  double r6inv = r2inv*r2inv*r2inv;
  double phi = r6inv*(lj3[itype][jtype]*r6inv-lj4[itype][jtype]) -
    offset[itype][jtype];
  return phi;
}
*/

/* ---------------------------------------------------------------------- */

/*
double PairLJExpand::fsingle(double r, int itype, int jtype)
{
  double rshift = r - shift[itype][jtype];
  double rshiftsq = rshift*rshift;
  double r2inv = 1.0/rshiftsq;
  double r6inv = r2inv*r2inv*r2inv;
  double force = r6inv * (lj1[itype][jtype]*r6inv - lj2[itype][jtype]);
  return force/rshift;
}
*/
