/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "set_type.h"
#include "atom.h"
#include "group.h"
#include "comm.h"
#include "neighbor.h"
#include "force.h"
#include "pair.h"
#include "random_park.h"
#include "error.h"

#define ATOM     1
#define BOND     2
#define ANGLE    3
#define DIHEDRAL 4
#define IMPROPER 5
#define DIPOLE   6

/* ---------------------------------------------------------------------- */

void SetType::set(int narg, char **arg)
{
  if (atom->natoms == 0)
    error->all("Set_type command with no atoms existing");

  if (comm->me == 0 && screen) fprintf(screen,"Setting types ...\n");

  if (narg != 3) error->all("Illegal set_type command");

  // identify group

  int igroup = group->find(arg[0]);
  if (igroup == -1) error->all("Cannot find set_type group ID");
  int groupbit = group->bitmask[igroup];
  
  // set style and ivalue

  int style;
  if (strcmp(arg[1],"atom") == 0) style = ATOM;
  else if (strcmp(arg[1],"bond") == 0) style = BOND;
  else if (strcmp(arg[1],"angle") == 0) style = ANGLE;
  else if (strcmp(arg[1],"dihedral") == 0) style = DIHEDRAL;
  else if (strcmp(arg[1],"improper") == 0) style = IMPROPER;
  else if (strcmp(arg[1],"dipole") == 0) style = DIPOLE;
  else error->all("Illegal set_type command");

  int ivalue = atoi(arg[2]);

  // consistency checks

  if ((style == BOND && atom->bonds_allow == 0) ||
      (style == ANGLE && atom->angles_allow == 0) ||
      (style == DIHEDRAL && atom->dihedrals_allow == 0) ||
      (style == IMPROPER && atom->impropers_allow == 0) ||
      (style == DIPOLE && atom->dipole_require == 0))
    error->all("Cannot perform set_type of disallowed interactions");

  if (style == ATOM && (ivalue < 1 || ivalue > atom->ntypes))
    error->all("Illegal type in set_type command");
  if (style == BOND && (ivalue < 1 || ivalue > atom->nbondtypes))
    error->all("Illegal type in set_type command");
  if (style == ANGLE && (ivalue < 1 || ivalue > atom->nangletypes))
    error->all("Illegal type in set_type command");
  if (style == DIHEDRAL && (ivalue < 1 || ivalue > atom->ndihedraltypes))
    error->all("Illegal type in set_type command");
  if (style == IMPROPER && (ivalue < 1 || ivalue > atom->nimpropertypes))
    error->all("Illegal type in set_type command");
  if (style == DIPOLE && ivalue <= 0)
    error->all("Illegal random # in set_type command");

  // border swap to insure type and mask is current for off-proc atoms
  // not needed for style ATOM or DIPOLE
  // must first set neighbor->cutneigh so comm will work
  // this requires pair->cutforce be set, do it via pair->init

  if (style != ATOM && style != DIPOLE) {
    if (force->pair == NULL) neighbor->cutneigh = neighbor->skin;
    else {
      force->pair->init();
      neighbor->cutneigh = force->pair->cutforce + neighbor->skin;
    }
    comm->setup();
    comm->exchange();
    comm->borders();
  }

  // change types
  // for style ATOM, atom must be in group
  // for style BOND (ANGLE, etc), each atom in interaction must be in group
  // for style DIPOLE, set each atom with dipole moment to random orientation

  int *mask = atom->mask;
  int *map = atom->map;
  int nlocal = atom->nlocal;
  int m,current,count,atom1,atom2,atom3,atom4;

  count = 0;

  if (style == ATOM)
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit) {
	current = atom->type[i];
	atom->type[i] = ivalue;
	if (current != ivalue) count++;
      }

  if (style == BOND)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_bond[i]; m++) {
	atom1 = map[atom->bond_atom[i][m]];
	if (atom1 == -1) error->one("Bond atom missing in set_type");
	if (mask[i] & groupbit && mask[atom1] & groupbit) {
	  current = atom->bond_type[i][m];
	  atom->bond_type[i][m] = ivalue;
	  if (current != ivalue) count++;
	}
      }

  if (style == ANGLE)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_angle[i]; m++) {
	atom1 = map[atom->angle_atom1[i][m]];
	atom2 = map[atom->angle_atom2[i][m]];
	atom3 = map[atom->angle_atom3[i][m]];
	if (atom1 == -1 || atom2 == -1 || atom3 == -1)
	  error->one("Angle atom missing in set_type");
	if (mask[atom1] & groupbit && mask[atom2] & groupbit &&
	    mask[atom3] & groupbit) {
	  current = atom->angle_type[i][m];
	  atom->angle_type[i][m] = ivalue;
	  if (current != ivalue) count++;
	}
      }

  if (style == DIHEDRAL)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_dihedral[i]; m++) {
	atom1 = map[atom->dihedral_atom1[i][m]];
	atom2 = map[atom->dihedral_atom2[i][m]];
	atom3 = map[atom->dihedral_atom3[i][m]];
	atom4 = map[atom->dihedral_atom4[i][m]];
	if (atom1 == -1 || atom2 == -1 || atom3 == -1 || atom4 == -1)
	  error->one("Dihedral atom missing in set_type");
	if (mask[atom1] & groupbit && mask[atom2] & groupbit &&
	    mask[atom3] & groupbit && mask[atom4] & groupbit) {
	  current = atom->dihedral_type[i][m];
	  atom->dihedral_type[i][m] = ivalue;
	  if (current != ivalue) count++;
	}
      }

  if (style == IMPROPER)
    for (int i = 0; i < nlocal; i++)
      for (m = 0; m < atom->num_improper[i]; m++) {
	atom1 = map[atom->improper_atom1[i][m]];
	atom2 = map[atom->improper_atom2[i][m]];
	atom3 = map[atom->improper_atom3[i][m]];
	atom4 = map[atom->improper_atom4[i][m]];
	if (atom1 == -1 || atom2 == -1 || atom3 == -1 || atom4 == -1)
	  error->one("Improper atom missing in set_type");
	if (mask[atom1] & groupbit && mask[atom2] & groupbit &&
	    mask[atom3] & groupbit && mask[atom4] & groupbit) {
	  current = atom->improper_type[i][m];
	  atom->improper_type[i][m] = ivalue;
	  if (current != ivalue) count++;
	}
      }

  if (style == DIPOLE) {
    double msq,scale;
    RanPark *random = new RanPark(ivalue);
    int *type = atom->type;
    double *dipole = atom->dipole;
    double **mu = atom->mu;
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit && dipole[type[i]] > 0.0) {
	mu[i][0] = random->uniform() - 0.5;
	mu[i][1] = random->uniform() - 0.5;
	mu[i][2] = random->uniform() - 0.5;
	msq = mu[i][0]*mu[i][0] + mu[i][1]*mu[i][1] + mu[i][2]*mu[i][2];
	scale = dipole[type[i]]/sqrt(msq);
	mu[i][0] *= scale;
	mu[i][1] *= scale;
	mu[i][2] *= scale;
	count++;
      }
    delete random;
  }

  if (comm->me == 0) {
    if (screen) fprintf(screen,"  %d changes made\n",count);
    if (logfile) fprintf(logfile,"  %d changes made\n",count);
  }
}
