/* -----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------ */

// Convert a LAMMPS binary restart file into an ASCII text data file
//
// this serial code must be compiled on a platform that can read the binary
//   restart file since binary formats are not compatible across all platforms
//
// Syntax: restart2data restart-file data-file
//
// each LAMMPS atom, pair, bond, angle, dihedral, improper style
//   must be added to this file in appropriate places

#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

// ---------------------------------------------------------------------
// Data class to hold problem
// ---------------------------------------------------------------------

class Data {
 public:

  // global settings

  char *version;
  int ntimestep;
  int nprocs;
  char *unit_style;
  int dimension;
  int px,py,pz;
  int newton_pair,newton_bond;
  int xperiodic,yperiodic,zperiodic;
  int boundary[3][2];
  char *atom_style;
  int natoms,nbonds,nangles,ndihedrals,nimpropers;
  int ntypes,nbondtypes,nangletypes,ndihedraltypes,nimpropertypes;
  int bond_per_atom,angle_per_atom,dihedral_per_atom,improper_per_atom;
  int bonds_allow,angles_allow,dihedrals_allow,impropers_allow;
  int mass_require,dipole_require;
  double xlo,xhi,ylo,yhi,zlo,zhi;
  double special_lj[4],special_coul[4];

  // force fields

  char *pair_style,*bond_style,*angle_style,*dihedral_style,*improper_style;

  double *pair_epsilon;
  double *pair_sigma;
  double *pair_shift;
  double *pair_eps14;
  double *pair_sigma14;
  double *pair_start;
  double *pair_stop;

  double *bond_harmonic_k;
  double *bond_harmonic_r0;

  double *bond_fene_k;
  double *bond_fene_r0;
  double *bond_fene_epsilon;
  double *bond_fene_sigma;
  double *bond_fene_shift;

  double *bond_nonlinear_epsilon;
  double *bond_nonlinear_r0;
  double *bond_nonlinear_lamda;

  double *angle_harmonic_k;
  double *angle_harmonic_theta0;

  double *angle_cosine_k;

  double *angle_charmm_k;
  double *angle_charmm_theta0;
  double *angle_charmm_k_ub;
  double *angle_charmm_r_ub;

  double *dihedral_harmonic_k;
  int *dihedral_harmonic_multiplicity;
  int *dihedral_harmonic_sign;

  double *dihedral_multi_a1;
  double *dihedral_multi_a2;
  double *dihedral_multi_a3;
  double *dihedral_multi_a4;
  double *dihedral_multi_a5;

  double *dihedral_charmm_k;
  int *dihedral_charmm_multiplicity;
  int *dihedral_charmm_sign;
  double *dihedral_charmm_weight;

  double *improper_harmonic_k;
  double *improper_harmonic_chi;

  double *improper_cvff_k;
  int *improper_cvff_sign;
  int *improper_cvff_multiplicity;

  // atom quantities

  int iatoms,ibonds,iangles,idihedrals,iimpropers;

  double *mass,*dipole;
  double *x,*y,*z,*vx,*vy,*vz;
  double *phix,*phiy,*phiz,*phivx,*phivy,*phivz;
  int *tag,*type,*mask,*image;
  int *molecule;
  double *q,*radius,*density;
  int *bond_type,*angle_type,*dihedral_type,*improper_type;
  int *bond_atom1,*bond_atom2;
  int *angle_atom1,*angle_atom2,*angle_atom3;
  int *dihedral_atom1,*dihedral_atom2,*dihedral_atom3,*dihedral_atom4;
  int *improper_atom1,*improper_atom2,*improper_atom3,*improper_atom4;
  
  // functions

  Data();
  void stats();
  void write(FILE *);
};

// ---------------------------------------------------------------------
// function prototypes
// ---------------------------------------------------------------------

void header(FILE *, Data &);
void groups(FILE *);
void masses(FILE *, Data &);
void dipoles(FILE *, Data &);
void force_fields(FILE *, Data &);
void modify(FILE *);
void pair(FILE *fp, Data &data);
void bond(FILE *fp, Data &data);
void angle(FILE *fp, Data &data);
void dihedral(FILE *fp, Data &data);
void improper(FILE *fp, Data &data);

int atom_atomic(double *, Data &);
int atom_bond(double *, Data &);
int atom_angle(double *, Data &);
int atom_full(double *, Data &);
int atom_eam(double *, Data &);
int atom_granular(double *, Data &);

typedef int (*FnPtr)(double *, Data &);
FnPtr atom;

int read_int(FILE *fp);
double read_double(FILE *fp);
char *read_char(FILE *fp);

// ---------------------------------------------------------------------
// main program
// ---------------------------------------------------------------------

main (int argc, char **argv)
{
  // syntax error check

  if (argc != 3) {
    printf("Syntax: restart2data restart-file data-file\n");
    exit(1);
  }
  
  // open restart file

  printf("Reading restart file ...\n");
  FILE *fp = fopen(argv[1],"rb");
  if (fp == NULL) {
    printf("ERROR: Cannot open restart file %s\n",argv[1]);
    exit(1);
  }

  // read beginning of restart file

  Data data;

  header(fp,data);
  groups(fp);
  if (data.mass_require) masses(fp,data);
  if (data.dipole_require) dipoles(fp,data);
  force_fields(fp,data);
  modify(fp);

  // set fn ptr to correct atom unpack routine

  if (strcmp(data.atom_style,"atomic") == 0) atom = &atom_atomic;
  else if (strcmp(data.atom_style,"bond") == 0) atom = &atom_bond;
  else if (strcmp(data.atom_style,"angle") == 0) atom = &atom_angle;
  else if (strcmp(data.atom_style,"full") == 0) atom = &atom_full;
  else if (strcmp(data.atom_style,"eam") == 0) atom = &atom_eam;
  else if (strcmp(data.atom_style,"granular") == 0) atom = &atom_granular;
  else {
    printf("ERROR: Unknown atom style %s\n",data.atom_style);
    exit(1);
  }

  // read atoms from restart file

  double *buf = NULL;
  int n,m;
  int maxbuf = 0;
  data.iatoms = data.ibonds = data.iangles = 
    data.idihedrals = data.iimpropers = 0;

  for (int iproc = 0; iproc < data.nprocs; iproc++) {
    fread(&n,sizeof(int),1,fp);
    if (n > maxbuf) {
      maxbuf = n;
      delete [] buf;
      buf = new double[maxbuf];
    }

    fread(buf,sizeof(double),n,fp);

    m = 0;
    while (m < n) m += atom(&buf[m],data);
  }

  fclose(fp);

  // print out stats

  data.stats();

  // write out data file

  printf("Writing data file ...\n");
  fp = fopen(argv[2],"w");
  if (fp == NULL) {
    printf("ERROR: Cannot open data file %s\n",argv[2]);
    exit(1);
  }

  data.write(fp);
  fclose(fp);
}

// ---------------------------------------------------------------------
// read header of restart file
// ---------------------------------------------------------------------

void header(FILE *fp, Data &data)
{
  char *version = "17 Jan 2005";

  int flag;
  flag = read_int(fp);

  while (flag >= 0) {

    if (flag == 0) {
      data.version = read_char(fp);
      if (strcmp(version,data.version) != 0) {
	char *str = "Restart file version does not match restart2data version";
	printf("WARNING %s\n",str);
	printf("  restart2data version = %s\n",version);
      }
    }
    else if (flag == 1) data.unit_style = read_char(fp);
    else if (flag == 2) data.ntimestep = read_int(fp);
    else if (flag == 3) data.dimension = read_int(fp);
    else if (flag == 4) data.nprocs = read_int(fp);
    else if (flag == 5) data.px = read_int(fp);
    else if (flag == 6) data.py = read_int(fp);
    else if (flag == 7) data.pz = read_int(fp);
    else if (flag == 8) data.newton_pair = read_int(fp);
    else if (flag == 9) data.newton_bond = read_int(fp);
    else if (flag == 10) data.xperiodic = read_int(fp);
    else if (flag == 11) data.yperiodic = read_int(fp);
    else if (flag == 12) data.zperiodic = read_int(fp);
    else if (flag == 13) data.boundary[0][0] = read_int(fp);
    else if (flag == 14) data.boundary[0][1] = read_int(fp);
    else if (flag == 15) data.boundary[1][0] = read_int(fp);
    else if (flag == 16) data.boundary[1][1] = read_int(fp);
    else if (flag == 17) data.boundary[2][0] = read_int(fp);
    else if (flag == 18) data.boundary[2][1] = read_int(fp);
    else if (flag == 19) data.atom_style = read_char(fp);
    else if (flag == 20) data.natoms = read_int(fp);
    else if (flag == 21) data.ntypes = read_int(fp);
    else if (flag == 22) data.nbonds = read_int(fp);
    else if (flag == 23) data.nbondtypes = read_int(fp);
    else if (flag == 24) data.bond_per_atom = read_int(fp);
    else if (flag == 25) data.nangles = read_int(fp);
    else if (flag == 26) data.nangletypes = read_int(fp);
    else if (flag == 27) data.angle_per_atom = read_int(fp);
    else if (flag == 28) data.ndihedrals = read_int(fp);
    else if (flag == 29) data.ndihedraltypes = read_int(fp);
    else if (flag == 30) data.dihedral_per_atom = read_int(fp);
    else if (flag == 31) data.nimpropers = read_int(fp);
    else if (flag == 32) data.nimpropertypes = read_int(fp);
    else if (flag == 33) data.improper_per_atom = read_int(fp);
    else if (flag == 34) data.xlo = read_double(fp);
    else if (flag == 35) data.xhi = read_double(fp);
    else if (flag == 36) data.ylo = read_double(fp);
    else if (flag == 37) data.yhi = read_double(fp);
    else if (flag == 38) data.zlo = read_double(fp);
    else if (flag == 39) data.zhi = read_double(fp);
    else if (flag == 40) data.special_lj[1] = read_double(fp);
    else if (flag == 41) data.special_lj[2] = read_double(fp);
    else if (flag == 42) data.special_lj[3] = read_double(fp);
    else if (flag == 43) data.special_coul[1] = read_double(fp);
    else if (flag == 44) data.special_coul[2] = read_double(fp);
    else if (flag == 45) data.special_coul[3] = read_double(fp);
    else {
      printf("ERROR: Invalid flag in header of restart file\n");
      exit(1);
    }

    flag = read_int(fp);
  }

  // set data flags based on atom_style

  if (strcmp(data.atom_style,"atomic") == 0) {
    data.bonds_allow = data.angles_allow = 
      data.dihedrals_allow = data.impropers_allow = 0;
    data.mass_require = 1;
    data.dipole_require = 0;
  } else if (strcmp(data.atom_style,"bond") == 0) {
    data.bonds_allow = 1;
    data.angles_allow = data.dihedrals_allow = data.impropers_allow = 0;
    data.mass_require = 1;
    data.dipole_require = 0;
  } else if (strcmp(data.atom_style,"angle") == 0) {
    data.bonds_allow = data.angles_allow = 1;
    data.dihedrals_allow = data.impropers_allow = 0;
    data.mass_require = 1;
    data.dipole_require = 0;
  } else if (strcmp(data.atom_style,"full") == 0) {
    data.bonds_allow = data.angles_allow = 
      data.dihedrals_allow = data.impropers_allow = 1;
    data.mass_require = 1;
    data.dipole_require = 0;
  } else if (strcmp(data.atom_style,"eam") == 0) {
    data.bonds_allow = data.angles_allow = 
      data.dihedrals_allow = data.impropers_allow = 0;
    data.mass_require = 1;
    data.dipole_require = 0;
  } else if (strcmp(data.atom_style,"granular") == 0) {
    data.bonds_allow = data.angles_allow = 
      data.dihedrals_allow = data.impropers_allow = 0;
    data.mass_require = 0;
    data.dipole_require = 0;
  } else {
    printf("ERROR: Unknown atom style %s\n",data.atom_style);
    exit(1);
  }
}

// ---------------------------------------------------------------------
// read group info from restart file, just ignore it
// ---------------------------------------------------------------------

void groups(FILE *fp)
{
  int ngroup;
  fread(&ngroup,sizeof(int),1,fp);

  int n;
  char *name;

  for (int i = 0; i < ngroup; i++) {
    name = read_char(fp);
    delete [] name;
  }
}

// ---------------------------------------------------------------------
// read masses from restart file
// ---------------------------------------------------------------------

void masses(FILE *fp, Data &data)
{
  data.mass = new double[data.ntypes+1];
  fread(&data.mass[1],sizeof(double),data.ntypes,fp);
}

// ---------------------------------------------------------------------
// read dipoles from restart file
// ---------------------------------------------------------------------

void dipoles(FILE *fp, Data &data)
{
  data.dipole = new double[data.ntypes+1];
  fread(&data.dipole[1],sizeof(double),data.ntypes,fp);
}

// ---------------------------------------------------------------------
// read force-field info from restart file
// ---------------------------------------------------------------------

void force_fields(FILE *fp, Data &data)
{
  int n;

  data.pair_style = read_char(fp);
  if (data.pair_style) pair(fp,data);

  if (data.bonds_allow) {
    data.bond_style = read_char(fp);
    if (data.bond_style) bond(fp,data);
  }
  if (data.angles_allow) {
    data.angle_style = read_char(fp);
    if (data.angle_style) angle(fp,data);
  }
  if (data.dihedrals_allow) {
    data.dihedral_style = read_char(fp);
    if (data.dihedral_style) dihedral(fp,data);
  }
  if (data.impropers_allow) {
    data.improper_style = read_char(fp);
    if (data.improper_style) improper(fp,data);
  }
}

// ---------------------------------------------------------------------
// read fix info from restart file, just ignore it
// ---------------------------------------------------------------------

void modify(FILE *fp)
{
  char *buf;
  int n;

  // nfix = # of fix entries with state

  int nfix;
  fread(&nfix,sizeof(int),1,fp);

  // read each entry with id string, style string, chunk of data

  for (int i = 0; i < nfix; i++) {
    buf = read_char(fp); delete [] buf;
    buf = read_char(fp); delete [] buf;
    buf = read_char(fp); delete [] buf;
  }

  // nfix = # of fix entries with peratom info

  int nfix_peratom;
  fread(&nfix_peratom,sizeof(int),1,fp);

  // read each entry with id string, style string, maxsize of one atom data

  for (int i = 0; i < nfix_peratom; i++) {
    buf = read_char(fp); delete [] buf;
    buf = read_char(fp); delete [] buf;
    n = read_int(fp);
  }
}

// ---------------------------------------------------------------------
// read atom info from restart file and store in data struct
// one method for each atom style
// ---------------------------------------------------------------------

int atom_atomic(double *buf, Data &data)
{
  if (data.iatoms == 0) {
    data.x = new double[data.natoms];
    data.y = new double[data.natoms];
    data.z = new double[data.natoms];
    data.vx = new double[data.natoms];
    data.vy = new double[data.natoms];
    data.vz = new double[data.natoms];
    data.tag = new int[data.natoms];
    data.type = new int[data.natoms];
    data.mask = new int[data.natoms];
    data.image = new int[data.natoms];
  }

  int iatoms = data.iatoms;

  int m = 1;
  data.x[iatoms] = buf[m++];
  data.y[iatoms] = buf[m++];
  data.z[iatoms] = buf[m++];
  data.vx[iatoms] = buf[m++];
  data.vy[iatoms] = buf[m++];
  data.vz[iatoms] = buf[m++];
  data.tag[iatoms] = static_cast<int> (buf[m++]);
  data.type[iatoms] = static_cast<int> (buf[m++]);
  data.mask[iatoms] = static_cast<int> (buf[m++]);
  data.image[iatoms] = static_cast<int> (buf[m++]);

  data.iatoms++;
  m = static_cast<int> (buf[0]);
  return m;
}

int atom_bond(double *buf, Data &data)
{
  if (data.iatoms == 0) {
    data.x = new double[data.natoms];
    data.y = new double[data.natoms];
    data.z = new double[data.natoms];
    data.tag = new int[data.natoms];
    data.type = new int[data.natoms];
    data.mask = new int[data.natoms];
    data.image = new int[data.natoms];
    data.vx = new double[data.natoms];
    data.vy = new double[data.natoms];
    data.vz = new double[data.natoms];

    data.molecule = new int[data.natoms];

    data.bond_type = new int[data.nbonds];
    data.bond_atom1 = new int[data.nbonds];
    data.bond_atom2 = new int[data.nbonds];
  }

  int iatoms = data.iatoms;

  int m = 1;
  data.x[iatoms] = buf[m++];
  data.y[iatoms] = buf[m++];
  data.z[iatoms] = buf[m++];
  data.tag[iatoms] = static_cast<int> (buf[m++]);
  data.type[iatoms] = static_cast<int> (buf[m++]);
  data.mask[iatoms] = static_cast<int> (buf[m++]);
  data.image[iatoms] = static_cast<int> (buf[m++]);
  data.vx[iatoms] = buf[m++];
  data.vy[iatoms] = buf[m++];
  data.vz[iatoms] = buf[m++];

  data.molecule[iatoms] = static_cast<int> (buf[m++]);

  int type,atom1,atom2;

  int n = static_cast<int> (buf[m++]);
  for (int k = 0; k < n; k++) {
    type = static_cast<int> (buf[m++]);
    atom1 = static_cast<int> (buf[m++]);
    if (data.newton_bond || data.tag[iatoms] == MIN(atom1,atom2)) {
      data.bond_type[data.ibonds] = type;
      data.bond_atom1[data.ibonds] = data.tag[iatoms];
      data.bond_atom2[data.ibonds] = atom1;
      data.ibonds++;
    }
  }

  data.iatoms++;
  m = static_cast<int> (buf[0]);
  return m;
}

int atom_angle(double *buf, Data &data)
{
  if (data.iatoms == 0) {
    data.x = new double[data.natoms];
    data.y = new double[data.natoms];
    data.z = new double[data.natoms];
    data.tag = new int[data.natoms];
    data.type = new int[data.natoms];
    data.mask = new int[data.natoms];
    data.image = new int[data.natoms];
    data.vx = new double[data.natoms];
    data.vy = new double[data.natoms];
    data.vz = new double[data.natoms];

    data.molecule = new int[data.natoms];

    data.bond_type = new int[data.nbonds];
    data.bond_atom1 = new int[data.nbonds];
    data.bond_atom2 = new int[data.nbonds];

    data.angle_type = new int[data.nangles];
    data.angle_atom1 = new int[data.nangles];
    data.angle_atom2 = new int[data.nangles];
    data.angle_atom3 = new int[data.nangles];
  }

  int iatoms = data.iatoms;

  int m = 1;
  data.x[iatoms] = buf[m++];
  data.y[iatoms] = buf[m++];
  data.z[iatoms] = buf[m++];
  data.tag[iatoms] = static_cast<int> (buf[m++]);
  data.type[iatoms] = static_cast<int> (buf[m++]);
  data.mask[iatoms] = static_cast<int> (buf[m++]);
  data.image[iatoms] = static_cast<int> (buf[m++]);
  data.vx[iatoms] = buf[m++];
  data.vy[iatoms] = buf[m++];
  data.vz[iatoms] = buf[m++];

  data.molecule[iatoms] = static_cast<int> (buf[m++]);

  int type,atom1,atom2,atom3;

  int n = static_cast<int> (buf[m++]);
  for (int k = 0; k < n; k++) {
    type = static_cast<int> (buf[m++]);
    atom1 = static_cast<int> (buf[m++]);
    if (data.newton_bond || data.tag[iatoms] == MIN(atom1,atom2)) {
      data.bond_type[data.ibonds] = type;
      data.bond_atom1[data.ibonds] = data.tag[iatoms];
      data.bond_atom2[data.ibonds] = atom1;
      data.ibonds++;
    }
  }

  n = static_cast<int> (buf[m++]);
  for (int k = 0; k < n; k++) {
    type = static_cast<int> (buf[m++]);
    atom1 = static_cast<int> (buf[m++]);
    atom2 = static_cast<int> (buf[m++]);
    atom3 = static_cast<int> (buf[m++]);
    if (data.newton_bond || data.tag[iatoms] == atom2) {
      data.angle_type[data.iangles] = type;
      data.angle_atom1[data.iangles] = atom1;
      data.angle_atom2[data.iangles] = atom2;
      data.angle_atom3[data.iangles] = atom3;
      data.iangles++;
    }
  }

  data.iatoms++;
  m = static_cast<int> (buf[0]);
  return m;
}

int atom_full(double *buf, Data &data)
{
  if (data.iatoms == 0) {
    data.x = new double[data.natoms];
    data.y = new double[data.natoms];
    data.z = new double[data.natoms];
    data.tag = new int[data.natoms];
    data.type = new int[data.natoms];
    data.mask = new int[data.natoms];
    data.image = new int[data.natoms];
    data.vx = new double[data.natoms];
    data.vy = new double[data.natoms];
    data.vz = new double[data.natoms];

    data.q = new double[data.natoms];
    data.molecule = new int[data.natoms];

    data.bond_type = new int[data.nbonds];
    data.bond_atom1 = new int[data.nbonds];
    data.bond_atom2 = new int[data.nbonds];

    data.angle_type = new int[data.nangles];
    data.angle_atom1 = new int[data.nangles];
    data.angle_atom2 = new int[data.nangles];
    data.angle_atom3 = new int[data.nangles];

    data.dihedral_type = new int[data.ndihedrals];
    data.dihedral_atom1 = new int[data.ndihedrals];
    data.dihedral_atom2 = new int[data.ndihedrals];
    data.dihedral_atom3 = new int[data.ndihedrals];
    data.dihedral_atom4 = new int[data.ndihedrals];

    data.improper_type = new int[data.nimpropers];
    data.improper_atom1 = new int[data.nimpropers];
    data.improper_atom2 = new int[data.nimpropers];
    data.improper_atom3 = new int[data.nimpropers];
    data.improper_atom4 = new int[data.nimpropers];
  }

  int iatoms = data.iatoms;

  int m = 1;
  data.x[iatoms] = buf[m++];
  data.y[iatoms] = buf[m++];
  data.z[iatoms] = buf[m++];
  data.tag[iatoms] = static_cast<int> (buf[m++]);
  data.type[iatoms] = static_cast<int> (buf[m++]);
  data.mask[iatoms] = static_cast<int> (buf[m++]);
  data.image[iatoms] = static_cast<int> (buf[m++]);
  data.vx[iatoms] = buf[m++];
  data.vy[iatoms] = buf[m++];
  data.vz[iatoms] = buf[m++];

  data.q[iatoms] = buf[m++];
  data.molecule[iatoms] = static_cast<int> (buf[m++]);

  int type,atom1,atom2,atom3,atom4;

  int n = static_cast<int> (buf[m++]);
  for (int k = 0; k < n; k++) {
    type = static_cast<int> (buf[m++]);
    atom1 = static_cast<int> (buf[m++]);
    if (data.newton_bond || data.tag[iatoms] < atom1) {
      data.bond_type[data.ibonds] = type;
      data.bond_atom1[data.ibonds] = data.tag[iatoms];
      data.bond_atom2[data.ibonds] = atom1;
      data.ibonds++;
    }
  }

  n = static_cast<int> (buf[m++]);
  for (int k = 0; k < n; k++) {
    type = static_cast<int> (buf[m++]);
    atom1 = static_cast<int> (buf[m++]);
    atom2 = static_cast<int> (buf[m++]);
    atom3 = static_cast<int> (buf[m++]);
    if (data.newton_bond || data.tag[iatoms] == atom2) {
      data.angle_type[data.iangles] = type;
      data.angle_atom1[data.iangles] = atom1;
      data.angle_atom2[data.iangles] = atom2;
      data.angle_atom3[data.iangles] = atom3;
      data.iangles++;
    }
  }

  n = static_cast<int> (buf[m++]);
  for (int k = 0; k < n; k++) {
    type = static_cast<int> (buf[m++]);
    atom1 = static_cast<int> (buf[m++]);
    atom2 = static_cast<int> (buf[m++]);
    atom3 = static_cast<int> (buf[m++]);
    atom4 = static_cast<int> (buf[m++]);
    if (data.newton_bond || data.tag[iatoms] == atom2) {
      data.dihedral_type[data.idihedrals] = type;
      data.dihedral_atom1[data.idihedrals] = atom1;
      data.dihedral_atom2[data.idihedrals] = atom2;
      data.dihedral_atom3[data.idihedrals] = atom3;
      data.dihedral_atom4[data.idihedrals] = atom4;
      data.idihedrals++;
    }
  }

  n = static_cast<int> (buf[m++]);
  for (int k = 0; k < n; k++) {
    type = static_cast<int> (buf[m++]);
    atom1 = static_cast<int> (buf[m++]);
    atom2 = static_cast<int> (buf[m++]);
    atom3 = static_cast<int> (buf[m++]);
    atom4 = static_cast<int> (buf[m++]);
    if (data.newton_bond || data.tag[iatoms] == atom2) {
      data.improper_type[data.iimpropers] = type;
      data.improper_atom1[data.iimpropers] = atom1;
      data.improper_atom2[data.iimpropers] = atom2;
      data.improper_atom3[data.iimpropers] = atom3;
      data.improper_atom4[data.iimpropers] = atom4;
      data.iimpropers++;
    }
  }

  data.iatoms++;
  m = static_cast<int> (buf[0]);
  return m;
}

int atom_eam(double *buf, Data &data)
{
  if (data.iatoms == 0) {
    data.x = new double[data.natoms];
    data.y = new double[data.natoms];
    data.z = new double[data.natoms];
    data.vx = new double[data.natoms];
    data.vy = new double[data.natoms];
    data.vz = new double[data.natoms];
    data.tag = new int[data.natoms];
    data.type = new int[data.natoms];
    data.mask = new int[data.natoms];
    data.image = new int[data.natoms];
  }

  int iatoms = data.iatoms;

  int m = 1;
  data.x[iatoms] = buf[m++];
  data.y[iatoms] = buf[m++];
  data.z[iatoms] = buf[m++];
  data.vx[iatoms] = buf[m++];
  data.vy[iatoms] = buf[m++];
  data.vz[iatoms] = buf[m++];
  data.tag[iatoms] = static_cast<int> (buf[m++]);
  data.type[iatoms] = static_cast<int> (buf[m++]);
  data.mask[iatoms] = static_cast<int> (buf[m++]);
  data.image[iatoms] = static_cast<int> (buf[m++]);

  data.iatoms++;
  m = static_cast<int> (buf[0]);
  return m;
}

int atom_granular(double *buf, Data &data)
{
  if (data.iatoms == 0) {
    data.x = new double[data.natoms];
    data.y = new double[data.natoms];
    data.z = new double[data.natoms];
    data.vx = new double[data.natoms];
    data.vy = new double[data.natoms];
    data.vz = new double[data.natoms];
    data.phix = new double[data.natoms];
    data.phiy = new double[data.natoms];
    data.phiz = new double[data.natoms];
    data.phivx = new double[data.natoms];
    data.phivy = new double[data.natoms];
    data.phivz = new double[data.natoms];
    data.radius = new double[data.natoms];
    data.density = new double[data.natoms];
    data.tag = new int[data.natoms];
    data.type = new int[data.natoms];
    data.mask = new int[data.natoms];
    data.image = new int[data.natoms];
  }

  int iatoms = data.iatoms;

  int m = 1;
  data.x[iatoms] = buf[m++];
  data.y[iatoms] = buf[m++];
  data.z[iatoms] = buf[m++];
  data.vx[iatoms] = buf[m++];
  data.vy[iatoms] = buf[m++];
  data.vz[iatoms] = buf[m++];
  data.phix[iatoms] = buf[m++];
  data.phiy[iatoms] = buf[m++];
  data.phiz[iatoms] = buf[m++];
  data.phivx[iatoms] = buf[m++];
  data.phivy[iatoms] = buf[m++];
  data.phivz[iatoms] = buf[m++];
  data.radius[iatoms] = buf[m++];
  data.density[iatoms] = buf[m++];
  data.tag[iatoms] = static_cast<int> (buf[m++]);
  data.type[iatoms] = static_cast<int> (buf[m++]);
  data.mask[iatoms] = static_cast<int> (buf[m++]);
  data.image[iatoms] = static_cast<int> (buf[m++]);

  data.iatoms++;
  m = static_cast<int> (buf[0]);
  return m;
}

// ---------------------------------------------------------------------
// pair coeffs
// one section for each pair style
// ---------------------------------------------------------------------

void pair(FILE *fp,Data &data)
{
  int i,j;
  double rtmp;
  int itmp;

  if (strcmp(data.pair_style,"lj/cut") == 0) {
    data.pair_epsilon = new double[data.ntypes+1];
    data.pair_sigma = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  if (i == j) {
	    fread(&data.pair_epsilon[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma[i],sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  } else {
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  }
	}
      }
  }

  else if (strcmp(data.pair_style,"lj/cut/coul/cut") == 0) {
    data.pair_epsilon = new double[data.ntypes+1];
    data.pair_sigma = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  if (i == j) {
	    fread(&data.pair_epsilon[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma[i],sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  } else {
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  }
	}
      }
  }

  else if (strcmp(data.pair_style,"lj/cut/coul/long") == 0) {
    data.pair_epsilon = new double[data.ntypes+1];
    data.pair_sigma = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  if (i == j) {
	    fread(&data.pair_epsilon[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma[i],sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  } else {
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  }
	}
      }
  }

  else if (strcmp(data.pair_style,"lj/cut/coul/debye") == 0) {
    data.pair_epsilon = new double[data.ntypes+1];
    data.pair_sigma = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  if (i == j) {
	    fread(&data.pair_epsilon[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma[i],sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  } else {
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  }
	}
      }
  }

  else if (strcmp(data.pair_style,"lj/expand") == 0) {
    data.pair_epsilon = new double[data.ntypes+1];
    data.pair_sigma = new double[data.ntypes+1];
    data.pair_shift = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  fread(&data.pair_epsilon[i],sizeof(double),1,fp);
	  fread(&data.pair_sigma[i],sizeof(double),1,fp);
	  fread(&data.pair_shift[i],sizeof(double),1,fp);
	  fread(&rtmp,sizeof(double),1,fp);
	} else {
	  fread(&rtmp,sizeof(double),1,fp);
	  fread(&rtmp,sizeof(double),1,fp);
	  fread(&rtmp,sizeof(double),1,fp);
	  fread(&rtmp,sizeof(double),1,fp);
	}
      }
  }

  else if (strcmp(data.pair_style,"soft") == 0) {
    data.pair_start = new double[data.ntypes+1];
    data.pair_stop = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  fread(&data.pair_start[i],sizeof(double),1,fp);
	  fread(&data.pair_stop[i],sizeof(double),1,fp);
	  fread(&rtmp,sizeof(double),1,fp);
	} else {
	  fread(&rtmp,sizeof(double),1,fp);
	  fread(&rtmp,sizeof(double),1,fp);
	  fread(&rtmp,sizeof(double),1,fp);
	}
      }
  }

  else if ((strcmp(data.pair_style,"lj/charmm/coul/charmm") == 0) ||
	   (strcmp(data.pair_style,"lj/charmm/coul/charmm/implicit") == 0)) {
    data.pair_epsilon = new double[data.ntypes+1];
    data.pair_sigma = new double[data.ntypes+1];
    data.pair_eps14 = new double[data.ntypes+1];
    data.pair_sigma14 = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  if (i == j) {
	    fread(&data.pair_epsilon[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma[i],sizeof(double),1,fp);
	    fread(&data.pair_eps14[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma14[i],sizeof(double),1,fp);
	  } else {
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  }
	}
      }
  }

  else if (strcmp(data.pair_style,"lj/charmm/coul/long") == 0) {
    data.pair_epsilon = new double[data.ntypes+1];
    data.pair_sigma = new double[data.ntypes+1];
    data.pair_eps14 = new double[data.ntypes+1];
    data.pair_sigma14 = new double[data.ntypes+1];

    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&rtmp,sizeof(double),1,fp);
    fread(&itmp,sizeof(int),1,fp);
    fread(&itmp,sizeof(int),1,fp);

    for (i = 1; i <= data.ntypes; i++)
      for (j = i; j <= data.ntypes; j++) {
	fread(&itmp,sizeof(int),1,fp);
	if (i == j && itmp == 0) {
	  printf("ERROR: Pair coeff %d,%d is not in restart file\n",i);
	  exit(1);
	}
	if (itmp) {
	  if (i == j) {
	    fread(&data.pair_epsilon[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma[i],sizeof(double),1,fp);
	    fread(&data.pair_eps14[i],sizeof(double),1,fp);
	    fread(&data.pair_sigma14[i],sizeof(double),1,fp);
	  } else {
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	    fread(&rtmp,sizeof(double),1,fp);
	  }
	}
      }
  }

  else if (strcmp(data.pair_style,"eam") == 0) {
  }

  else if ((strcmp(data.pair_style,"gran/history") == 0) ||
	   (strcmp(data.pair_style,"gran/no_history") == 0) ||
	   (strcmp(data.pair_style,"gran/hertzian") == 0)) {
    double xkk = read_double(fp);
    double gamman = read_double(fp);
    double xmu = read_double(fp);
    int dampflag = read_int(fp);
  }

  else {
    printf("ERROR: Unknown pair style %s\n",data.pair_style);
    exit(1);
  }
}

// ---------------------------------------------------------------------
// bond coeffs
// one section for each bond style
// ---------------------------------------------------------------------

void bond(FILE *fp, Data &data)
{
  if (strcmp(data.bond_style,"harmonic") == 0) {
    data.bond_harmonic_k = new double[data.nbondtypes+1];
    data.bond_harmonic_r0 = new double[data.nbondtypes+1];
    fread(&data.bond_harmonic_k[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_harmonic_r0[1],sizeof(double),data.nbondtypes,fp);
  }

  else if (strcmp(data.bond_style,"fene") == 0) {
    data.bond_fene_k = new double[data.nbondtypes+1];
    data.bond_fene_r0 = new double[data.nbondtypes+1];
    data.bond_fene_epsilon = new double[data.nbondtypes+1];
    data.bond_fene_sigma = new double[data.nbondtypes+1];
    fread(&data.bond_fene_k[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_fene_r0[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_fene_epsilon[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_fene_sigma[1],sizeof(double),data.nbondtypes,fp);
  }

  else if (strcmp(data.bond_style,"fene/expand") == 0) {
    data.bond_fene_k = new double[data.nbondtypes+1];
    data.bond_fene_r0 = new double[data.nbondtypes+1];
    data.bond_fene_epsilon = new double[data.nbondtypes+1];
    data.bond_fene_sigma = new double[data.nbondtypes+1];
    data.bond_fene_shift = new double[data.nbondtypes+1];
    fread(&data.bond_fene_k[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_fene_r0[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_fene_epsilon[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_fene_sigma[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_fene_shift[1],sizeof(double),data.nbondtypes,fp);
  }

  else if (strcmp(data.bond_style,"nonlinear") == 0) {
    data.bond_nonlinear_epsilon = new double[data.nbondtypes+1];
    data.bond_nonlinear_r0 = new double[data.nbondtypes+1];
    data.bond_nonlinear_lamda = new double[data.nbondtypes+1];
    fread(&data.bond_nonlinear_epsilon[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_nonlinear_r0[1],sizeof(double),data.nbondtypes,fp);
    fread(&data.bond_nonlinear_lamda[1],sizeof(double),data.nbondtypes,fp);
  }

  else {
    printf("ERROR: Unknown bond style %s\n",data.bond_style);
    exit(1);
  }
}

// ---------------------------------------------------------------------
// angle coeffs
// one section for each angle style
// ---------------------------------------------------------------------

void angle(FILE *fp, Data &data)
{
  if (strcmp(data.angle_style,"harmonic") == 0) {
    data.angle_harmonic_k = new double[data.nangletypes+1];
    data.angle_harmonic_theta0 = new double[data.nangletypes+1];
    fread(&data.angle_harmonic_k[1],sizeof(double),data.nangletypes,fp);
    fread(&data.angle_harmonic_theta0[1],sizeof(double),data.nangletypes,fp);
  }

  else if (strcmp(data.angle_style,"cosine") == 0) {
    data.angle_cosine_k = new double[data.nangletypes+1];
    fread(&data.angle_cosine_k[1],sizeof(double),data.nangletypes,fp);
  }

  else if (strcmp(data.angle_style,"charmm") == 0) {
    data.angle_charmm_k = new double[data.nangletypes+1];
    data.angle_charmm_theta0 = new double[data.nangletypes+1];
    data.angle_charmm_k_ub = new double[data.nangletypes+1];
    data.angle_charmm_r_ub = new double[data.nangletypes+1];
    fread(&data.angle_charmm_k[1],sizeof(double),data.nangletypes,fp);
    fread(&data.angle_charmm_theta0[1],sizeof(double),data.nangletypes,fp);
    fread(&data.angle_charmm_k_ub[1],sizeof(double),data.nangletypes,fp);
    fread(&data.angle_charmm_r_ub[1],sizeof(double),data.nangletypes,fp);
  }

  else {
    printf("ERROR: Unknown angle style %s\n",data.angle_style);
    exit(1);
  }
}

// ---------------------------------------------------------------------
// dihedral coeffs
// one section for each dihedral style
// ---------------------------------------------------------------------

void dihedral(FILE *fp, Data &data)
{
  if (strcmp(data.dihedral_style,"harmonic") == 0) {
    data.dihedral_harmonic_k = new double[data.ndihedraltypes+1];
    data.dihedral_harmonic_multiplicity = new int[data.ndihedraltypes+1];
    data.dihedral_harmonic_sign = new int[data.ndihedraltypes+1];
    fread(&data.dihedral_harmonic_k[1],sizeof(double),data.ndihedraltypes,fp);
    fread(&data.dihedral_harmonic_multiplicity[1],sizeof(int),
	  data.ndihedraltypes,fp);
    fread(&data.dihedral_harmonic_sign[1],sizeof(int),data.ndihedraltypes,fp);
  }

  else if (strcmp(data.dihedral_style,"multi/harmonic") == 0) {
    data.dihedral_multi_a1 = new double[data.ndihedraltypes+1];
    data.dihedral_multi_a2 = new double[data.ndihedraltypes+1];
    data.dihedral_multi_a3 = new double[data.ndihedraltypes+1];
    data.dihedral_multi_a4 = new double[data.ndihedraltypes+1];
    data.dihedral_multi_a5 = new double[data.ndihedraltypes+1];
    fread(&data.dihedral_multi_a1[1],sizeof(double),data.ndihedraltypes,fp);
    fread(&data.dihedral_multi_a2[1],sizeof(double),data.ndihedraltypes,fp);
    fread(&data.dihedral_multi_a3[1],sizeof(double),data.ndihedraltypes,fp);
    fread(&data.dihedral_multi_a4[1],sizeof(double),data.ndihedraltypes,fp);
    fread(&data.dihedral_multi_a5[1],sizeof(double),data.ndihedraltypes,fp);
  }

  else if (strcmp(data.dihedral_style,"charmm") == 0) {
    data.dihedral_charmm_k = new double[data.ndihedraltypes+1];
    data.dihedral_charmm_multiplicity = new int[data.ndihedraltypes+1];
    data.dihedral_charmm_sign = new int[data.ndihedraltypes+1];
    data.dihedral_charmm_weight = new double[data.ndihedraltypes+1];
    fread(&data.dihedral_charmm_k[1],sizeof(double),data.ndihedraltypes,fp);
    fread(&data.dihedral_charmm_multiplicity[1],sizeof(int),
	  data.ndihedraltypes,fp);
    fread(&data.dihedral_charmm_sign[1],sizeof(int),data.ndihedraltypes,fp);
    fread(&data.dihedral_charmm_weight[1],sizeof(double),
	  data.ndihedraltypes,fp);
  }

  else {
    printf("ERROR: Unknown dihedral style %s\n",data.dihedral_style);
    exit(1);
  }
}

// ---------------------------------------------------------------------
// improper coeffs
// one section for each improper style
// ---------------------------------------------------------------------

void improper(FILE *fp, Data &data)
{
  if (strcmp(data.improper_style,"harmonic") == 0) {
    data.improper_harmonic_k = new double[data.nimpropertypes+1];
    data.improper_harmonic_chi = new double[data.nimpropertypes+1];
    fread(&data.improper_harmonic_k[1],sizeof(double),data.nimpropertypes,fp);
    fread(&data.improper_harmonic_chi[1],sizeof(double),
	  data.nimpropertypes,fp);
  }

  else if (strcmp(data.improper_style,"cvff") == 0) {
    data.improper_cvff_k = new double[data.nimpropertypes+1];
    data.improper_cvff_sign = new int[data.nimpropertypes+1];
    data.improper_cvff_multiplicity = new int[data.nimpropertypes+1];
    fread(&data.improper_cvff_k[1],sizeof(double),data.nimpropertypes,fp);
    fread(&data.improper_cvff_sign[1],sizeof(int),data.nimpropertypes,fp);
    fread(&data.improper_cvff_multiplicity[1],sizeof(int),
	  data.nimpropertypes,fp);
  }

  else {
    printf("ERROR: Unknown improper style %s\n",data.improper_style);
    exit(1);
  }
}

// ---------------------------------------------------------------------
// initialize ptrs in Data
// ---------------------------------------------------------------------

Data::Data()
{
  pair_style = bond_style = angle_style =
    dihedral_style = improper_style = NULL;
}

// ---------------------------------------------------------------------
// print out stats on problem
// ---------------------------------------------------------------------

void Data::stats()
{
  printf("  Restart file version = %s\n",version);
  printf("  Ntimestep = %d\n",ntimestep);
  printf("  Nprocs = %d\n",nprocs);
  printf("  Natoms = %d\n",natoms);
  printf("  Nbonds = %d\n",nbonds);
  printf("  Nangles = %d\n",nangles);
  printf("  Ndihedrals = %d\n",ndihedrals);
  printf("  Nimpropers = %d\n",nimpropers);
  printf("  Unit style = %s\n",unit_style);
  printf("  Atom style = %s\n",atom_style);
  printf("  Pair style = %s\n",pair_style);
  printf("  Bond style = %s\n",bond_style);
  printf("  Angle style = %s\n",angle_style);
  printf("  Dihedral style = %s\n",dihedral_style);
  printf("  Improper style = %s\n",improper_style);
  printf("  Xlo/xhi = %g %g\n",xlo,xhi);
  printf("  Ylo/yhi = %g %g\n",ylo,yhi);
  printf("  Zlo/zhi = %g %g\n",zlo,zhi);
  printf("  Periodicity = %d %d %d\n",xperiodic,yperiodic,zperiodic);
  printf("  Boundary = %d %d, %d %d, %d %d\n",boundary[0][0],boundary[0][1],
	 boundary[1][0],boundary[1][1],boundary[2][0],boundary[2][1]);
}

// ---------------------------------------------------------------------
// write the data file
// ---------------------------------------------------------------------

void Data::write(FILE *fp)
{
  fprintf(fp,"LAMMPS data file from restart file: timestep = %d, procs = %d\n",
	  ntimestep,nprocs);
  
  fprintf(fp,"\n");

  fprintf(fp,"%d atoms\n",natoms);
  if (nbonds) fprintf(fp,"%d bonds\n",nbonds);
  if (nangles) fprintf(fp,"%d angles\n",nangles);
  if (ndihedrals) fprintf(fp,"%d dihedrals\n",ndihedrals);
  if (nimpropers) fprintf(fp,"%d impropers\n",nimpropers);

  fprintf(fp,"\n");

  fprintf(fp,"%d atom types\n",ntypes);
  if (nbondtypes) fprintf(fp,"%d bond types\n",nbondtypes);
  if (nangletypes) fprintf(fp,"%d angle types\n",nangletypes);
  if (ndihedraltypes) fprintf(fp,"%d dihedral types\n",ndihedraltypes);
  if (nimpropertypes) fprintf(fp,"%d improper types\n",nimpropertypes);

  fprintf(fp,"\n");
  
  fprintf(fp,"%g %g xlo xhi\n",xlo,xhi);
  fprintf(fp,"%g %g ylo yhi\n",ylo,yhi);
  fprintf(fp,"%g %g zlo zhi\n",zlo,zhi);

  if (mass_require) {
    fprintf(fp,"\nMasses\n\n");
    for (int i = 1; i <= ntypes; i++) fprintf(fp,"%d %g\n",i,mass[i]);
  }

  if (dipole_require) {
    fprintf(fp,"\nDipoles\n\n");
    for (int i = 1; i <= ntypes; i++) fprintf(fp,"%d %g\n",i,dipole[i]);
  }
  
  if (pair_style) {
    if ((strcmp(pair_style,"lj/cut") == 0) || 
	(strcmp(pair_style,"lj/cut/coul/cut") == 0) ||
	(strcmp(pair_style,"lj/cut/coul/long") == 0) ||
	(strcmp(pair_style,"lj/cut/coul/debye") == 0)) {
      fprintf(fp,"\nPair Coeffs\n\n");
      for (int i = 1; i <= ntypes; i++)
	fprintf(fp,"%d %g %g\n",i,
		pair_epsilon[i],pair_sigma[i]);
    } else if (strcmp(pair_style,"lj/expand") == 0) {
      fprintf(fp,"\nPair Coeffs\n\n");
      for (int i = 1; i <= ntypes; i++)
	fprintf(fp,"%d %g %g %g\n",i,
		pair_epsilon[i],pair_sigma[i],
		pair_shift[i]);
    } else if (strcmp(pair_style,"soft") == 0) {
      fprintf(fp,"\nPair Coeffs\n\n");
      for (int i = 1; i <= ntypes; i++)
	fprintf(fp,"%d %g %g %g\n",i,
		pair_start[i],pair_stop[i]);
    } else if ((strcmp(pair_style,"lj/charmm/coul/charmm") == 0) ||
	       (strcmp(pair_style,"lj/charmm/coul/long") == 0) ||
	       (strcmp(pair_style,"lj/charmm/coul/charmm/implicit") == 0)) {
      fprintf(fp,"\nPair Coeffs\n\n");
      for (int i = 1; i <= ntypes; i++)
	fprintf(fp,"%d %g %g %g %g\n",i,
		pair_epsilon[i],pair_sigma[i],
		pair_eps14[i],pair_sigma14[i]);
    } else if (strcmp(pair_style,"eam") == 0) {
    } else if (strcmp(pair_style,"gran/history") == 0) {
    } else if (strcmp(pair_style,"gran/no_history") == 0) {
    } else if (strcmp(pair_style,"gran/hertzian") == 0) {
    } else {
      printf("ERROR: Unknown pair style %s\n",pair_style);
      exit(1);
    }
  }

  if (bond_style) {
    fprintf(fp,"\nBond Coeffs\n\n");
    if (strcmp(bond_style,"harmonic") == 0)
      for (int i = 1; i <= nbondtypes; i++)
	fprintf(fp,"%d %g %g\n",i,
		bond_harmonic_k[i],bond_harmonic_r0[i]);
    else if (strcmp(bond_style,"fene") == 0)
      for (int i = 1; i <= nbondtypes; i++)
	fprintf(fp,"%d %g %g %g %g\n",i,
		bond_fene_k[i],bond_fene_r0[i],
		bond_fene_epsilon[i],bond_fene_sigma[i]);
    else if (strcmp(bond_style,"fene/expand") == 0)
      for (int i = 1; i <= nbondtypes; i++)
	fprintf(fp,"%d %g %g %g %g %g\n",i,
		bond_fene_k[i],bond_fene_r0[i],
		bond_fene_epsilon[i],bond_fene_sigma[i],bond_fene_shift[i]);
    else if (strcmp(bond_style,"nonlinear") == 0)
      for (int i = 1; i <= nbondtypes; i++)
	fprintf(fp,"%d %g %g %g\n",i,
		bond_nonlinear_epsilon[i],bond_nonlinear_r0[i],
		bond_nonlinear_lamda[i]);
    else {
      printf("ERROR: Unknown bond style %s\n",bond_style);
      exit(1);
    }
  }

  if (angle_style) {
    double PI = 3.1415926;           // convert back to degrees

    fprintf(fp,"\nAngle Coeffs\n\n");
    if (strcmp(angle_style,"harmonic") == 0)
      for (int i = 1; i <= nangletypes; i++)
	fprintf(fp,"%d %g %g\n",i,
		angle_harmonic_k[i],angle_harmonic_theta0[i]/PI*180.0);
    else if (strcmp(angle_style,"cosine") == 0)
      for (int i = 1; i <= nangletypes; i++)
	fprintf(fp,"%d %g\n",i,
		angle_cosine_k[i]);
    else if (strcmp(angle_style,"charmm") == 0)
      for (int i = 1; i <= nangletypes; i++)
	fprintf(fp,"%d %g %g %g %g\n",i,
		angle_charmm_k[i],angle_charmm_theta0[i]/PI*180.0,
		angle_charmm_k_ub[i],angle_charmm_r_ub[i]);
    else {
      printf("ERROR: Unknown angle style %s\n",angle_style);
      exit(1);
    }
  }

  if (dihedral_style) {
    fprintf(fp,"\nDihedral Coeffs\n\n");
    if (strcmp(dihedral_style,"harmonic") == 0)
      for (int i = 1; i <= ndihedraltypes; i++)
	fprintf(fp,"%d %g %d %d\n",i,
		dihedral_harmonic_k[i],dihedral_harmonic_multiplicity[i],
		dihedral_harmonic_sign[i]);
    else if (strcmp(dihedral_style,"multi/harmonic") == 0)
      for (int i = 1; i <= ndihedraltypes; i++)
	fprintf(fp,"%d %g %g %g %g %g\n",i,
		dihedral_multi_a1[i],dihedral_multi_a2[i],
		dihedral_multi_a3[i],dihedral_multi_a4[i],
		dihedral_multi_a5[i]);
    else if (strcmp(dihedral_style,"charmm") == 0)
      for (int i = 1; i <= ndihedraltypes; i++)
	fprintf(fp,"%d %g %d %d %g\n",i,
		dihedral_charmm_k[i],dihedral_charmm_multiplicity[i],
		dihedral_charmm_sign[i],dihedral_charmm_weight[i]);
    else {
      printf("ERROR: Unknown dihedral style %s\n",dihedral_style);
      exit(1);
    }
  }

  if (improper_style) {
    fprintf(fp,"\nImproper Coeffs\n\n");
    if (strcmp(improper_style,"harmonic") == 0)
      for (int i = 1; i <= nimpropertypes; i++)
	fprintf(fp,"%d %g %g\n",i,
		improper_harmonic_k[i],improper_harmonic_chi[i]);
    else if (strcmp(improper_style,"cvff") == 0)
      for (int i = 1; i <= nimpropertypes; i++)
	fprintf(fp,"%d %g %d %d\n",i,
		improper_cvff_k[i],improper_cvff_sign[i],
		improper_cvff_multiplicity[i]);
    else {
      printf("ERROR: Unknown improper style %s\n",improper_style);
      exit(1);
    }
  }

  if (natoms) {
    fprintf(fp,"\nAtoms\n\n");

    int ix,iy,iz;
    for (int i = 0; i < natoms; i++) {
      ix = image[i]%1000 - 500;
      iy = (image[i]/1000)%1000 - 500;
      iz = image[i]/1000000 - 500;

      if (strcmp(atom_style,"atomic") == 0)
	fprintf(fp,"%d %d %g %g %g %d %d %d\n",
		tag[i],type[i],x[i],y[i],z[i],ix,iy,iz);
      else if (strcmp(atom_style,"bond") == 0)
	fprintf(fp,"%d %d %d %g %g %g %d %d %d\n",
		tag[i],molecule[i],type[i],x[i],y[i],z[i],ix,iy,iz);
      else if (strcmp(atom_style,"angle") == 0)
	fprintf(fp,"%d %d %d %g %g %g %d %d %d\n",
		tag[i],molecule[i],type[i],x[i],y[i],z[i],ix,iy,iz);
      else if (strcmp(atom_style,"full") == 0)
	fprintf(fp,"%d %d %d %g %g %g %g %d %d %d\n",
		tag[i],molecule[i],type[i],q[i],x[i],y[i],z[i],ix,iy,iz);
      else if (strcmp(atom_style,"eam") == 0)
	fprintf(fp,"%d %d %g %g %g %d %d %d\n",
		tag[i],type[i],x[i],y[i],z[i],ix,iy,iz);
      else if (strcmp(atom_style,"granular") == 0)
	fprintf(fp,"%d %d %g %g %g %g %g %d %d %d\n",
		tag[i],type[i],2.0*radius[i],density[i],
		x[i],y[i],z[i],ix,iy,iz);
      else {
	printf("ERROR: Unknown atom style %s\n",atom_style);
	exit(1);
      }
    }
  }

  if (natoms) {
    fprintf(fp,"\nVelocities\n\n");
    for (int i = 0; i < natoms; i++)
      if (strcmp(atom_style,"granular") == 0)
	fprintf(fp,"%d %g %g %g %g %g %g\n",tag[i],vx[i],vy[i],vz[i],
		phivx[i],phivy[i],phivz[i]);
      else
	fprintf(fp,"%d %g %g %g\n",tag[i],vx[i],vy[i],vz[i]);
  }

  if (nbonds) {
    fprintf(fp,"\nBonds\n\n");
    for (int i = 0; i < nbonds; i++)
      fprintf(fp,"%d %d %d %d\n",
	      i+1,bond_type[i],bond_atom1[i],bond_atom2[i]);
  }

  if (nangles) {
    fprintf(fp,"\nAngles\n\n");
    for (int i = 0; i < nangles; i++)
      fprintf(fp,"%d %d %d %d %d\n",
	      i+1,angle_type[i],angle_atom1[i],angle_atom2[i],angle_atom3[i]);
  }

  if (ndihedrals) {
    fprintf(fp,"\nDihedrals\n\n");
    for (int i = 0; i < ndihedrals; i++)
      fprintf(fp,"%d %d %d %d %d %d\n",
	      i+1,dihedral_type[i],dihedral_atom1[i],dihedral_atom2[i],
	      dihedral_atom3[i],dihedral_atom4[i]);
  }

  if (nimpropers) {
    fprintf(fp,"\nImpropers\n\n");
    for (int i = 0; i < nimpropers; i++)
      fprintf(fp,"%d %d %d %d %d %d\n",
	      i+1,improper_type[i],improper_atom1[i],improper_atom2[i],
	      improper_atom3[i],improper_atom4[i]);
  }
}

/* read an int from restart file */

int read_int(FILE *fp)
{
  int value;
  fread(&value,sizeof(int),1,fp);
  return value;
}

/* read a double from restart file */

double read_double(FILE *fp)
{
  double value;
  fread(&value,sizeof(double),1,fp);
  return value;
}

/* read a char str from restart file */

char *read_char(FILE *fp)
{
  int n;
  fread(&n,sizeof(int),1,fp);
  if (n == 0) return NULL;
  char *value = new char[n];
  fread(value,sizeof(char),n,fp);
  return value;
}
